/*     
 * $RCSfile: psync_internal.h,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: psync_internal.h,v $
 * Revision 1.1  1994/04/18 22:02:08  slm
 * Initial revision
 *
 */

/*
 * #define BROADCAST           1
 * #define FREE_STABLE_MESSAGE 1
 */

#define MAX_ORPHANS 50		/* (2*MAX_PART) */
#define MAX_CONV    101		/* max conversations at the same time */
#define MAX_NODES   500		/* max # nodes per conversation */
#define NUM_NODES   4000

#define FIRST       1		/*     message types    */
#define REGULAR     2		/*                      */
#define RETRANSMIT  3		/*                      */
#define REPLY       4		/*                      */
#define RESTART     5		/*                      */

#define SAME_PSYNChost(h1,h2) \
    ((h1).d == (h2).d && (h1).c == (h2).c && \
     (h1).b == (h2).b && (h1).a == (h2).a ? 1 : 0)

#define SAME_MID(h1,h2) \
    (((h1).host.d == (h2).host.d && (h1).host.c == (h2).host.c && \
      (h1).host.b == (h2).host.b && (h1).host.a == (h2).host.a && \
      (h1).msg_id == (h2).msg_id) ? 1 : 0)

#define MAX(a,b) ((a) > (b) ? (a) : (b))


typedef unsigned long PART_MASK;	/* must have >= MAX_PART bits */

typedef struct {	/* conversation id */
    IPhost   host;			/* IP host */
    unsigned conv_id;			/* conversation id */
} CID;

typedef struct {	/* message id */
    IPhost   host;			/* IP host */
    unsigned msg_id;			/* message id */
} MID;

typedef struct node {	/* a node in the context graph */
    MID           mid;			/* message id */
    Msg           *msg;			/* message */
    int           sender;		/* sender's id */
    char          recvd;		/* whether received (NOT USED) */
    PART_MASK     stable_part_mask;	/* bit i is on if, for stability, a msg
					 * needs to be received from part i */
    unsigned char num_prev, num_next;	/* number of parents and children */
    struct node   *prev[MAX_PART];	/* list of parents */
    struct node   *next[MAX_PART];	/* list of children */
} NODE;

typedef struct {	/* PSYNC protocol state */
    IPhost    host;			/* IP host */
    unsigned  conv_id;			/* conversation id */
    Map       map_openenable;		/* <localport,iphost> --> <hlp> */
    Map       map_conv;			/* <CID> --> <psync_sessn> */
    Semaphore Ipsessn;			/* for opening IP session */
    Semaphore Ihgraph;			/* for accessing Ih (context graph) */
} PSYNCstate;

typedef struct  {	/* PSYNC session state */
    char           first_msg;			/* set for first message */
    char           loopback;			/* set for loopback */
    NODE           *root;			/* root of Ih (context graph) */
    int            mypartnum;
    PSYNCaddr      psyncaddr;
    CID            cid;				/* conversation id */
    unsigned       msg_id;			/* message id */
    unsigned char  num_part;			/* # participants */
    unsigned char  num_leaves;			/* # leaves in the view */
    unsigned char  num_orphans;			/* # orphans in orphan_list */
    PSYNCaddr      part_list[MAX_PART];		/* list of participants */
    NODE           *leaves[MAX_PART];		/* list of leaf nodes */
    Msg            *orphan_list[MAX_ORPHANS];	/* list of orphan messages */
    NODE           *lastmsg[MAX_PART];		/* last msg by the sender */
    MID            pushing_mid;			/* used in PSYNC_ACKMESSAGE */
    XObj           rec_sessn;			/* used in PSYNC_RESTART & */
    int            rec_opcode;			/*     PSYNC_SETRECOVERY   */
    struct rcv_msg {
	NODE           *node;
	struct rcv_msg *next;
    }              *next_msg, *last_msg;	/* used in PSYNC_OUTSTANDING */
    PART_MASK      down_part_mask;		/* bit i is on if participant
						 * i is down */
    PART_MASK      ignore_part_mask;		/* bit i is on if msg from
					         * part i is to be ignored */
    Map            map_tree;			/* <iphost,mid> --> node> */
    Semaphore      Pushing;			/* for pushing */
} Sstate;

typedef struct {	/* first message header */
    char          type;
    CID           cid;
    MID           mid;
    int           sender;
    unsigned char num_part;
    PSYNCaddr     part_list[MAX_PART];
} PSYNCheader1;

typedef struct {	/* regular message header */
    char          type;
    CID           cid;
    MID           mid;
    int           sender;
    unsigned char num_depend;
    MID           depend_list[MAX_PART];
} PSYNCheader2;

typedef struct {	/* common message header (for peeking) */
    char type;
    CID  cid;
} PSYNCheader3;

typedef struct {	/* arg structure for rev_hdr_store and rec_own_msg */
    XObj   psync_sessn;
    Sstate *state;
    Msg    *msg;
    MID    mid;
} ARGstruct;

typedef struct {	/* used to compute indegree for nodes */
    NODE *node;
    int  indegree;
} INDEGREE;

/* Q for participants invited to the conversation but not yet listening */
typedef struct invtQ {
    CID          cid;
    PSYNCport    port;
    Msg          *msg;
    XObj         lls;
    struct invtQ *next;
} INVITEQ;
