/* 
 * $RCSfile: host_list.h,v $
 *
 * handles the host_list data type and related operations
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * $Log: host_list.h,v $
 * Revision 1.1  1995/10/17 22:24:02  ho
 * Initial revision
 *
 * Revision 1.2  1995/04/28  00:21:08  dorgival
 * Minor syntax changes to make the IRIX compiler happy
 *
 * Revision 1.1  1995/04/24  18:48:15  dorgival
 * Initial revision
 *
 * Revision 1.1  1995/04/05  17:52:37  dorgival
 * Initial revision
 *
 *
 */

#define LOCALHOSTTYPE       IPhost

#define GRPMEM_MAP_SZ		16
#define GRPMEM_MAX_PARTS	16
#define GRPMEM_MAX_PORTSIZE	16

typedef struct {
  int count;
  LOCALHOSTTYPE* hosts;
} host_list;

#define HOST_LIST_SIZE(hl)   ((hl)->count)
#define HOST_LIST_HOSTS(hl)  ((hl)->hosts)
#define HOST_LIST_HOST(hl,idx) (&((hl)->hosts[idx]))
#define HOST_LIST_ADD(hl,idx,host) (bcopy((char*)host, (char*)HOST_LIST_HOST(hl,idx),sizeof(LOCALHOSTTYPE)))

#define isalpha( c )  (((c)==' ')||((c)=='\n')||((c)=='\t'))

#define same_host( h1,h2 ) ((h1==h2)|(h1&&h2&&!bcmp((char*)h1,(char*)h2,sizeof(LOCALHOSTTYPE))))

typedef int (*foreach_func_t( LOCALHOSTTYPE*, void*, void* ));

#ifdef __STDC__

void            host_list_init(    host_list* hl );
void            host_list_clear(   host_list* hl );
void            host_list_destroy( host_list* hl );
void            host_list_add(     host_list* hl, LOCALHOSTTYPE* h );
void            host_list_del(     host_list* hl, LOCALHOSTTYPE* h );
LOCALHOSTTYPE*  host_list_find(    host_list* hl, LOCALHOSTTYPE* h );
void            host_list_foreach( host_list* hl, foreach_func_t f,
				   void* arg1, void* arg2 );

#else

void host_list_add( );
void host_list_del( );
LOCALHOSTTYPE*  host_list_find( );
void host_list_foreach( );
void host_list_clear( );

#endif


#define add_group_member( ss, h ) host_list_add( SS_GROUP_MEMBERS(ss), h )
#define del_group_member( ss, h ) host_list_del( SS_GROUP_MEMBERS(ss), h )
#define find_group_member( ss, h ) host_list_find( SS_GROUP_MEMBERS(ss), h )
#define foreach_group_member( ss, f, arg1, arg2 )                \
	host_list_foreach( SS_GROUP_MEMBERS(ss), f, arg1, arg2 )

#define add_group_owner( ss, h ) host_list_add( SS_GROUP_OWNER(ss), h )
#define del_group_owner( ss, h ) host_list_del( SS_GROUP_OWNER(ss), h )
#define find_group_owner( ss, h ) host_list_find( SS_GROUP_OWNER(ss), h )
#define foreach_group_owner( ss, f, arg1, arg2 )                \
	host_list_foreach( SS_GROUP_OWNER(ss), f, arg1, arg2 )

#define add_group_admin( ss, h )  host_list_add( SS_GROUP_ADMINS(ss), h )
#define del_group_admin( ss, h )  host_list_del( SS_GROUP_ADMINS(ss), h )
#define find_group_admin( ss, h ) host_list_find( SS_GROUP_ADMINS(ss), h )
#define foreach_group_admin( ss, f, arg1, arg2 )                \
	host_list_foreach( SS_GROUP_ADMINS(ss), f, arg1, arg2 )
