/*     
 * $RCSfile: ipcp.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994  Arizona Board of Regents
 *
 * $Log: ipcp.c,v $
 * Revision 1.3  1994/10/28 20:35:28  andrey
 * *** empty log message ***
 *
 * Revision 1.2  1994/10/27  17:26:00  andrey
 * removed non-ansi definitions
 *
 * Revision 1.1  1994/10/24  20:48:44  andrey
 * Initial revision
 *
 *
 * $Revision: 1.3 $
 * $Date: 1994/10/28 20:35:28 $
 */

/*
 * IP Control Protocol
 * for PPP
 * 
 * uses skeleton FSM and negotiable options routines from PPP
 */

int     traceipcpp;

#include "xkernel.h"
#include "ppp.h"
#include "ipcp.h"
#include "ipcp_constants.h"

#include "../ppp/ppp_options.h"
#include "../ppp/ppp_fsm.h"

#include "ipcp_i.h"

static void             getproc_protl(XObj);
static void             getproc_sessn(XObj);
extern xkern_return_t   ipcpProcessOpts(XObj);
static xkern_return_t   ipcpCloseProtl( XObj );
static xkern_return_t   ipcpCloseSessn( XObj );
static int              ipcpControlProtl( XObj, int, char *, int ) ;
static int              ipcpControlSessn( XObj, int, char *, int ) ;
static XObj             ipcpCreateSessn( XObj, XObj, XObj, XObj );
static xkern_return_t   ipcpDemux( XObj, XObj, Msg * );
static XObj             ipcpOpen( XObj, XObj, XObj, Part * );
static xkern_return_t   ipcpOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t   ipcpOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t   ipcpOpenDisableAll( XObj, XObj );

static xkern_return_t   ipcpOpenDone( XObj, XObj, XObj, XObj);
static xkern_return_t   ipcpCloseDone( XObj );

static xkern_return_t   ipcpPop( XObj, XObj, Msg *, void * );
/*static xmsg_handle_t    ipcpPush( XObj, Msg * );	*/
static xkern_return_t	process_message(XObj, Msg *);
static void		SSTATE_init(SSTATE *, XObj);

#include	"ipcp_fsm_table.h"

#include	"ipcp_options.c"
#include	"ipcp_config.c"

void
ipcp_init(XObj self)
{
  PSTATE * pstate = X_NEW(PSTATE);
  XObj     llp;

  PSTATE_init(pstate);
  xTraceP0(self, TR_GROSS_EVENTS, "init");

  getproc_protl(self);
  self->state = (void *)pstate;
  (void)ipcpProcessOpts(self);

  /* should ensure that lower protocol is actually PPP */
  if ( ! xIsProtocol(llp = xGetDown(self, 0)) ) {
    Kabort("IPCP could not get lower protocol");
  }

  if ( xOpenEnable(self, self, xGetDown(self, 0), 0) == XK_FAILURE) {
    xTraceP0(self, TR_ALWAYS, "init: can't openenable lower protocol");
    FREE(pstate);
    self->state = 0;
    return;
  }
  xTraceP0(self, TR_EVENTS, "init done");
}

static void
getproc_protl(XObj self)
{
  xAssert(xIsProtocol(self));
  self->close = ipcpCloseProtl;
  self->control = ipcpControlProtl;
  self->open = ipcpOpen;
  self->opendone = ipcpOpenDone;
  self->openenable = ipcpOpenEnable;
  self->opendisable = ipcpOpenDisable;
  self->closedone = ipcpCloseDone;
  self->demux = ipcpDemux;
  self->opendisableall = ipcpOpenDisableAll;
}

static void
getproc_sessn(XObj self)
{
  xAssert(xIsSession(self));
  self->pop = ipcpPop;
  self->control = ipcpControlSessn;
  self->close = ipcpCloseSessn;
}

static xkern_return_t
ipcpCloseProtl(XObj self)
{
  xTraceP0(self, TR_EVENTS, "CloseProtl");
  return XK_FAILURE;
}

/*
 * should be via lower protocol
 */
static xkern_return_t
ipcpCloseSessn(XObj self)
{
  xTraceP0(self, TR_EVENTS, "CloseSessn");
  return XK_FAILURE;
}

static int
ipcpControlProtl(XObj self, int opcode, char * buf, int len)
{
  PSTATE * pstate = (PSTATE *)self->state;
  xTraceP0(self, TR_EVENTS, "ControlProtl");

  switch(opcode) {

    case PPP_NEGOTIATE:
      return ipcpControlSessn(pstate->TheSession, opcode, buf, len);
    break;

  }
  return 0;
}

static int
ipcpControlSessn(XObj self, int opcode, char * buf, int len)
{
  SSTATE * sstate = (SSTATE *) self->state;
  FSM_info * fi = sstate->ipcp_fsm;

  xTraceP0(self, TR_EVENTS, "ControlSessn");

  switch(opcode) {

    case PPP_NEGOTIATE:
      if (len < sizeof(sstate->my_agreed_options))
        return 0;
      ppp_process_event(fi, F_open);
      fi->waiting_on_nego = 1;
      semWait(&fi->Nego_sem);
      if (fi->nego_ok) {
        bcopy(buf, (char *)sstate->my_agreed_options,
                  sizeof(sstate->my_agreed_options));
        return sizeof(sstate->my_agreed_options);
      }
      else
        return 0;
    break;

  }
  return 0;
}

/*
 * always the result of an OpenDone
 */
static XObj
ipcpCreateSessn(XObj self, XObj hlpRcv, XObj hlpType, XObj lls)
{
  XObj s;
  SSTATE * sstate = (SSTATE *)X_NEW(SSTATE);

  xTraceP0(self, TR_EVENTS, "CreateSessn");
  s = xCreateSessn(getproc_sessn, hlpRcv, hlpType, self, 1, &lls);
  s->state = sstate;
  SSTATE_init(sstate, s);

  sstate->my_nego_options = X_NEW(PPP_options);
  PPP_options_init(sstate->my_nego_options);
  ppp_copy_options(sstate->my_nego_options,
           &((PSTATE *)self->state)->rom_ipcp_options,
           ipcp_duplicate_option);

  xIfTraceP(s, TR_DETAILED) {
    printf("session options\n");
    ppp_print_options(sstate->my_nego_options, ipcp_print_opt);
    printf("\n");
  }
  return s;
}

/*
 * Corresponds to a DOWN event
 */
static xkern_return_t
ipcpCloseDone(XObj self)
{
  FSM_info * fi = ((SSTATE *)self->state)->ipcp_fsm;

  xTraceP0(self, TR_EVENTS, "CloseDone");
  xSetDown(self, 0, ERR_XOBJ);
  ppp_process_event(fi, F_down);
  return XK_SUCCESS;
}

/*
 * Corresponds to an UP event
 */
static xkern_return_t
ipcpOpenDone(XObj self, XObj lls, XObj llp, XObj hlpType)
{
  PSTATE * pstate = (PSTATE *)self->state;
  FSM_info * fi;

  xTraceP0(self, TR_EVENTS, "OpenDone");
  if (pstate->TheSession == ERR_XOBJ) {
    pstate->TheSession = ipcpCreateSessn(self, self, self, lls);
    xDuplicate(pstate->TheSession);
  }
  else {
    xSetDown(pstate->TheSession, 0, lls);
  }
  fi = ((SSTATE*)pstate->TheSession->state)->ipcp_fsm;
  ppp_process_event(fi, F_up);
  return XK_SUCCESS;
}

/*
 * may need to inject an OPEN event into IPCP's automaton
 */
static xkern_return_t
ipcpDemux(XObj self, XObj lls, Msg * msg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  FSM_info * fi = ((SSTATE *)pstate->TheSession->state)->ipcp_fsm;

  xTraceP0(self, TR_EVENTS, "Demux");

  switch (fi->state) {
    case F_initial:
    break;
    case F_starting:
    break;
    case F_closed:
      ppp_process_event(fi, F_open);
    break;
    case F_stopped:
    break;
    case F_closing:
    break;
    case F_stopping:
    break;

    default:
    break;
  }
  return xPop(pstate->TheSession, lls, msg, 0);
}

static XObj
ipcpOpen(XObj self, XObj hlpRcv, XObj hlpType, Part * p)
{
  xTraceP0(self, TR_EVENTS, "Open");
  return ERR_XOBJ;
}

static xkern_return_t
ipcpOpenDisable( XObj self, XObj hlpRcv, XObj hlpType, Part * p)
{
  xTraceP0(self, TR_EVENTS, "OpenDisable");
  return XK_FAILURE;
}

static xkern_return_t
ipcpOpenEnable(XObj self, XObj hlpRcv, XObj hlpType, Part * p)
{
  xTraceP0(self, TR_EVENTS, "OpenEnable");
  return XK_FAILURE;
}

static xkern_return_t
ipcpOpenDisableAll(XObj self, XObj hlpRcv)
{
  xTraceP0(self, TR_EVENTS, "OpenDisableAll");
  return XK_FAILURE;
}

static xkern_return_t
ipcpPop(XObj self, XObj ds, Msg * msg, void * hdr)
{
  xTraceP0(self, TR_EVENTS, "Pop");

  /* had better be no arg, and no ds */
  return process_message(self, msg);
}

/* ------------------------------------------------------------------------ */
/*                          FSM routines                                    */
/* ------------------------------------------------------------------------ */

static void
signal_negotiation_done(XObj sess)
{
  SSTATE * sstate = (SSTATE *) sess->state;
  FSM_info * fi = sstate->ipcp_fsm;

  if (fi->nego_ok)
    xControl(xGetDown(sess, 0), PPP_DONE_NEGOTIATION,
           (void *)&sstate->my_agreed_options,
            sizeof(&sstate->my_agreed_options));
  else
    xControl(xGetDown(sess, 0), PPP_DONE_NEGOTIATION, 0, 0);
}

static Msg *
make_config_req_msg(XObj sess)
{
  SSTATE * sstate = (SSTATE *)sess->state;

  return  make_config_msg(sstate->my_nego_options);
}

static Msg *
make_config_ack_msg(XObj sess)
{
  SSTATE * sstate = (SSTATE *)sess->state;

  return  make_config_msg(sstate->peer_agreed_options);
}

static Msg *
make_config_nak_msg(XObj sess)
{
  SSTATE * sstate = (SSTATE *)sess->state;

  return  make_config_msg(sstate->peer_nego_options);
}

static Msg *
make_config_rej_msg(XObj sess)
{
  SSTATE * sstate = (SSTATE *)sess->state;

  return  make_config_msg(sstate->peer_un_nego_options);
}

/*
 * pull apart message, then turn crank on automaton
 */
static xkern_return_t
process_message(XObj sess, Msg * msg)
{
  int code, len, id;
  fsm_event next_event = NO_EVENT;
  SSTATE * sstate = (SSTATE *)sess->state;
  FSM_info * fi = sstate->ipcp_fsm;
  Msg * msg_copy = X_NEW(Msg);

  xTraceP0(sess, TR_DETAILED, "process message");
  /* destroyed after action is performed above */
  msgConstructCopy(msg_copy, msg);

  code = ppp_decode_packet(msg, &id, &len);

  xTraceP3(sess, TR_DETAILED, "process message type %x id %x len %d",
          code, id, len);

  if (msgLen(msg) + sizeof(PPP_header) != len) {
    xTraceP2(sess, TR_SOFT_ERRORS, "length %d at odds with msg length %d",
            len, msgLen(msg) + sizeof(PPP_header) );

/*
 * just discard
    return;
 */
  }

  len -= sizeof(PPP_header);

  switch(code) {

    /* generate either F_rcv_cfg_req_good or F_rcv_cfg_req_bad */
    case PPP_CONF_REQ:
      {
        PPP_options * lo;

        lo = make_config_options(msg, len);

        xIfTraceP(sess, TR_DETAILED) {
          printf("Received options:\n");
          ppp_print_options(lo, ipcp_print_opt);
          printf("\n");
        }
        if (options_acceptable(sess, lo,
                               &sstate->peer_un_nego_options,
                               &sstate->peer_nego_options, 1) == XK_FAILURE) {
          if (sstate->peer_un_nego_options) {
            xIfTraceP(sess, TR_DETAILED) {
              printf("I rejected options:\n");
              ppp_print_options(sstate->peer_un_nego_options, ipcp_print_opt);
              printf("\n");
            }
            fi->nak_not_rej = 0;
          }
          else {
            xIfTraceP(sess, TR_DETAILED) {
              printf("I nak'ed options:\n");
              ppp_print_options(sstate->peer_nego_options, ipcp_print_opt);
              printf("\n");
            }
            fi->nak_not_rej = 1;
          }
          next_event = F_rcv_cfg_req_bad;
        }
        else {
          xTraceP0(sess, TR_DETAILED, "options ok");
          sstate->peer_agreed_options = lo;
          next_event = F_rcv_cfg_req_good;
        }
      }
    break;

    case PPP_CONF_ACK:		/* generate F_rcv_cfg_ack */
      {
        PPP_options * lo = make_config_options(msg, len);
        xIfTraceP(sess, TR_DETAILED) {
          printf("peer ack'ed options:\n");
          ppp_print_options(lo, ipcp_print_opt);
          printf("\n");
        }

        if (sstate->my_nego_options->ident != id) {
          xTraceP2(sess, TR_SOFT_ERRORS,
                  "questionable nack/rej - sent id %d, received %d",
                  sstate->my_nego_options->ident, id);
          /* ignore packet
             return;
*/
        }
        if (!ppp_options_match(sstate->my_nego_options,
                               lo, ipcp_option_matches)) {
          xTraceP0(sess, TR_SOFT_ERRORS, "questionable ack - option mismatch");
          /* ignore packet
             return;
*/
        }
        if (sstate->my_agreed_options == 0)
          sstate->my_agreed_options = lo;
        else {
          ppp_options_append(sstate->my_agreed_options, lo);
        }
        next_event = F_rcv_cfg_ack;
      }
    break;

    case PPP_CONF_REJ:		/* generate F_rcv_cfg_nak */
      {

        PPP_options * new_opts, * lo;
        if (sstate->my_nego_options->ident != id) {
          xTraceP2(sess, TR_SOFT_ERRORS,
                  "questionable rej - sent id %d, received %d",
                  sstate->my_nego_options->ident, id);
          /* ignore packet
             return;
*/
        }
        lo = make_config_options(msg, len);
        xIfTraceP(sess, TR_DETAILED) {
          printf("peer rejected options:\n");
          ppp_print_options(lo, ipcp_print_opt);
          printf("\n");
        }

        /* retire rejected options */
        new_opts = ppp_option_differences(sstate->my_nego_options, lo,
                                          ipcp_option_matches,
                                          ipcp_duplicate_option);
        /* no options remain? */
        if (!new_opts) {
          new_opts = X_NEW(PPP_options);
        }
        new_opts->ident = id++;
        sstate->my_un_nego_options = lo;
        ppp_destroy_options(sstate->my_nego_options, ipcp_destroy_opt);
        sstate->my_nego_options = new_opts;
        next_event = F_rcv_cfg_nak;
      }
    break;

    case PPP_CONF_NAK:		/* generate F_rcv_cfg_nak */
      {
        PPP_options * new_opts, * lo;
        if (sstate->my_nego_options->ident != id) {
          xTraceP2(sess, TR_SOFT_ERRORS,
                  "questionable nack/rej - sent id %d, received %d",
                  sstate->my_nego_options->ident, id);
          /* ignore packet
             return;
*/
        }
        /* recalculate options */
        lo = make_config_options(msg, len);
        xIfTraceP(sess, TR_DETAILED) {
          printf("peer nak'ed options:\n");
          ppp_print_options(lo, ipcp_print_opt);
          printf("\n");
        }

        new_opts = IPCP_options_reconsider(sess, lo);
        new_opts->ident = id++;
        ppp_destroy_options(sstate->my_nego_options, ipcp_destroy_opt);
        sstate->my_nego_options = new_opts;
        next_event = F_rcv_cfg_nak;
      }
    break;

    case PPP_TERM_REQ:		/* generate F_rcv_term_req */
      next_event = F_rcv_term_req;
    break;

    case PPP_TERM_ACK:		/* generate F_rcv_term_ack */
      next_event = F_rcv_term_ack;
    break;

    default:			/* generate LCP_rcv_unk_code */
      next_event = F_rcv_unk_code;
      /* dump the code field */
      msgPopDiscard(msg_copy, 1);
      fi->msg = msg_copy;
    break;
  }
  ppp_process_event(fi, next_event);
  return XK_SUCCESS;
}

/* ------------------------------------------------------------------------ */

static void
FSM_info_init(FSM_info * fi, XObj sess, xkern_return_t (*pm)(XObj, Msg *),
  Msg * (*mcrm)(XObj),
  Msg * (*mcam)(XObj),
  Msg * (*mcnm)(XObj),
  Msg * (*mcjm)(XObj),
  void (*snd)(XObj))
{
  fi->session = sess;
  fi->state_table = &ipcp_state_table;
  fi->action_table = &ipcp_fsm_action;
  fi->timerev = 0;
  semInit(&fi->Nego_sem, 0);
  fi->state = F_initial;
  fi->msg = 0;
  fi->restart_counter = 0;
  fi->timerev = 0;
  fi->waiting_on_nego = 0;
  fi->process_message = pm;
  fi->make_config_req_msg = mcrm;
  fi->make_config_ack_msg = mcam;
  fi->make_config_nak_msg = mcnm;
  fi->make_config_rej_msg = mcjm;
  fi->signal_negotiation_done = snd;
}

/* ------------------------------------------------------------------------ */

static void
SSTATE_init(SSTATE * ssp, XObj sess)
{
  ssp->ipcp_fsm = X_NEW(FSM_info);
  FSM_info_init(ssp->ipcp_fsm, sess,
      process_message, make_config_req_msg,
      make_config_ack_msg, make_config_nak_msg,
      make_config_rej_msg, signal_negotiation_done);

  ssp->ipcp_fsm->time_options =
         &((PSTATE*)sess->myprotl->state)->implem_options;

  ssp->my_agreed_options = 0;
  ssp->my_nego_options = 0;
  ssp->my_un_nego_options = 0;
  ssp->peer_agreed_options = 0;
  ssp->peer_nego_options = 0;
  ssp->peer_un_nego_options = 0;
}

