/*
 * $RCSfile: ipcp_options.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994  Arizona Board of Regents
 *
 * $Log: ipcp_options.c,v $
 * Revision 1.3  1994/10/28 20:35:28  andrey
 * *** empty log message ***
 *
 * Revision 1.2  1994/10/27  17:26:00  andrey
 * removed non-ansi definitions
 *
 * Revision 1.1  1994/10/24  20:48:44  andrey
 * Initial revision
 *
 *
 * $Revision: 1.3 $
 * $Date: 1994/10/28 20:35:28 $
 */

/*
 * ipcp specific negotiation option specific routines for use with
 * ppp_option's list routines
 */

static void * ipcp_duplicate_option(void * this, u_char type);
static void ipcp_destroy_opt( void * this, u_char type);
static void ipcp_print_opt(void * this, u_char type);
static int ipcp_option_matches(void * from, void * to, u_char type);

static void *
ipcp_duplicate_option(void * this, u_char type)
{
  void * that = 0;

  switch(type) {
    case IPCP_ADDRESS:
      that = X_NEW(ADDRESS_opt);
      ((ADDRESS_opt*)that)->Host = ((ADDRESS_opt*)this)->Host;
    break;
    case IPCP_COMPRESSION:
      that = X_NEW(COMPRESSION_opt);
      ((COMPRESSION_opt*)that)->max_slot_id =
         ((COMPRESSION_opt*)this)->max_slot_id;
      ((COMPRESSION_opt*)that)->comp_slot_id =
         ((COMPRESSION_opt*)this)->comp_slot_id;
    break;

    case IPCP_UNKNOWN:
      that = X_NEW(UNKNOWN_opt);
      ((UNKNOWN_opt *)that)->type = ((UNKNOWN_opt *)this)->type;
      ((UNKNOWN_opt *)that)->length = ((UNKNOWN_opt *)this)->length;
      ((UNKNOWN_opt *)that)->buf = xMalloc(((UNKNOWN_opt *)this)->length);
      bcopy(((UNKNOWN_opt *)this)->buf, ((UNKNOWN_opt *)that)->buf,
            ((UNKNOWN_opt *)that)->length);
    break;
  }

  return that;
}

static void
ipcp_destroy_opt(void * this, u_char type)
{
  switch(type) {
    case IPCP_UNKNOWN:
      FREE(((UNKNOWN_opt *)this)->buf);

    case IPCP_ADDRESS:
    case IPCP_COMPRESSION:

      FREE(this);
    break;
  }
}

static void
ipcp_print_opt(void * this, u_char type)
{
  switch(type) {
    case IPCP_ADDRESS:
      printf("address %s\t", ipHostStr(&((ADDRESS_opt*)this)->Host) );
    break;
    case IPCP_COMPRESSION:
      printf("V-J compression max-slot %d comp-slot %d\t",
            ((COMPRESSION_opt*)this)->max_slot_id,
            ((COMPRESSION_opt*)this)->comp_slot_id);
    break;
    case IPCP_UNKNOWN:
      printf("unknown\t");
    break;
  }
}

static int
ipcp_option_matches(void * from, void * to, u_char type)
{
  switch(type) {
    case IPCP_ADDRESS:
      return IP_EQUAL(((ADDRESS_opt *)from)->Host, ((ADDRESS_opt *)to)->Host);
    break;
    case IPCP_COMPRESSION:
      return
         ((COMPRESSION_opt *)from)->max_slot_id ==
         ((COMPRESSION_opt *)to)->max_slot_id &&
         ((COMPRESSION_opt *)from)->comp_slot_id ==
         ((COMPRESSION_opt *)to)->comp_slot_id ;
    break;
    default:
      return 0;
  }
}

