/*
 * $RCSfile: ipcp_rom.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994  Arizona Board of Regents
 *
 * $Log: ipcp_rom.c,v $
 * Revision 1.2  1994/10/27 17:26:00  andrey
 * removed non-ansi definitions
 *
 * Revision 1.1  1994/10/24  20:48:44  andrey
 * Initial revision
 *
 *
 * $Revision: 1.2 $
 * $Date: 1994/10/27 17:26:00 $
 */

/*
 * romfile options processing for PPP's IP control protocol
 */

#include "xkernel.h"
#include "romopt.h"
#include "ipcp_constants.h"

#include "../ppp/ppp_fsm.h"
#include "../ppp/ppp_options.h"
#include "ipcp_i.h"

#include "ipcp_options.c"

static xkern_return_t   ipcp_rom_term_ret( XObj, char **, int, int, VOID * );
static xkern_return_t   ipcp_rom_conf_ret( XObj, char **, int, int, VOID * );
static xkern_return_t   ipcp_rom_fail_ret( XObj, char **, int, int, VOID * );
static xkern_return_t   ipcp_rom_timeout( XObj, char **, int, int, VOID * );
static xkern_return_t   ipcp_rom_compression( XObj, char **, int, int, VOID * );
static xkern_return_t   ipcp_rom_address( XObj, char **, int, int, VOID * );
xkern_return_t ipcpProcessOpts(XObj self);

/*
 * ipcp terminate_retries <num>
 * ipcp configure_retries <num>
 * ipcp failure_retries <num>
 * ipcp timeout <num>
 * ipcp address ipaddr
 * ipcp compression maxslotid compslotid
 */

static XObjRomOpt       rOpts[] =
{
  {  "terminate_retries", 3, ipcp_rom_term_ret },
  {  "configure_retries", 3, ipcp_rom_conf_ret },
  {  "failure_retries", 3, ipcp_rom_fail_ret },
  {  "timeout", 3, ipcp_rom_timeout },
  {  "address", 3, ipcp_rom_address },
  {  "that_address", 3, ipcp_rom_address },
  {  "compression", 4, ipcp_rom_compression },
};

xkern_return_t
ipcpProcessOpts(XObj self)
{
  return findXObjRomOpts(self, rOpts, sizeof(rOpts)/sizeof(XObjRomOpt), 0);
}

static xkern_return_t
ipcp_rom_term_ret(
  XObj        self,
  char        **strs,
  int         nFields,
  int line, 
  VOID        *arg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  int k;

  if (sscanf(strs[2], "%d", &k) != 1) {
    xTraceP2(self, TR_SOFT_ERRORS, "can't read rom %s entry: %s",
        strs[1], strs[2]);
    return XK_FAILURE;
  }
  pstate->implem_options.term_retry = k;
  return XK_SUCCESS;
}

static xkern_return_t
ipcp_rom_conf_ret(
    XObj        self,
    char        **strs,
    int         nFields,
    int 	line,
    VOID        *arg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  int k;

  if (sscanf(strs[2], "%d", &k) != 1) {
    xTraceP2(self, TR_SOFT_ERRORS, "can't read rom %s entry: %s",
        strs[1], strs[2]);
    return XK_FAILURE;
  }
  pstate->implem_options.config_retry = k;
  return XK_SUCCESS;
}

static xkern_return_t
ipcp_rom_fail_ret(
    XObj        self,
    char        **strs,
    int         nFields,
    int 	line,
    VOID        *arg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  int k;

  if (sscanf(strs[2], "%d", &k) != 1) {
    xTraceP2(self, TR_SOFT_ERRORS, "can't read rom %s entry: %s",
        strs[1], strs[2]);
    return XK_FAILURE;
  }
  pstate->implem_options.config_naks = k;
  return XK_SUCCESS;
}

static xkern_return_t
ipcp_rom_timeout(
    XObj        self,
    char        **strs,
    int         nFields,
    int		line,
    VOID        *arg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  int k;

  if (sscanf(strs[2], "%d", &k) != 1) {
    xTraceP2(self, TR_SOFT_ERRORS, "can't read rom %s entry: %s",
        strs[1], strs[2]);
    return XK_FAILURE;
  }
  pstate->implem_options.timeval = k;
  return XK_SUCCESS;
}


static xkern_return_t
ipcp_rom_address(
    XObj        self,
    char        **strs,
    int         nFields,
    int		line,
    VOID        *arg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  PPP_options * pp_o = &pstate->rom_ipcp_options;
  ADDRESS_opt * ao;
  IPhost ipHost;

  if (strcmp(strs[1], "that_address")) {
    if ( str2ipHost(&ipHost, strs[2]) == XK_FAILURE ) {
      xTraceP1(self, TR_SOFT_ERRORS, "can't read ipaddress %s", strs[2]);
      return XK_FAILURE;
    }
  }
  else {
    if ( str2ipHost(&ipHost, strs[2]) == XK_FAILURE ) {
      xTraceP1(self, TR_SOFT_ERRORS, "can't read ipaddress %s", strs[2]);
      return XK_FAILURE;
    }
  }
  ao = X_NEW(ADDRESS_opt);
  ao->Host = ipHost;
  ppp_add_option(pp_o, ao, IPCP_ADDRESS);
  return XK_SUCCESS;
}

static xkern_return_t
ipcp_rom_compression(
    XObj        self,
    char        **strs,
    int         nFields,
    int		line,
    VOID        *arg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  PPP_options * pp_o = &pstate->rom_ipcp_options;
  int maxslot, compslot;
  COMPRESSION_opt * co;

  if (sscanf(strs[2], "%d", &maxslot) != 1 ||
      sscanf(strs[3], "%d", &compslot) != 1) {

    xTraceP3(self, TR_SOFT_ERRORS, "can't read rom entry: %s %s %s",
        strs[1], strs[2], strs[3]);
    return XK_FAILURE;
  }

  if (maxslot < 0 || maxslot > 0xFF || compslot < 0 || compslot > 0x1) {
    xTraceP3(self, TR_SOFT_ERRORS, "rom entry out of bounds: %s %s %s",
        strs[1], strs[2], strs[3]);
    return XK_FAILURE;
  }

  co = X_NEW(COMPRESSION_opt);
  co->max_slot_id = maxslot;
  co->comp_slot_id = compslot;
  ppp_add_option(pp_o, co, IPCP_COMPRESSION);
  return XK_SUCCESS;
}

