/*     
 * $RCS$
 *
 * $LOG$
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994,1993,1991,1990  Arizona Board of Regents
 *
 *
 *
 */

#include "xkernel.h"
#include "km.h"
#include "dheckx.h"
#include "saidmgr.h"
#include "saidmgr_i.h"
#include "vnet.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>


#define MAX_PRESET_KEYS  (SAIDMGR_KEYSIZE+1)
#define NUMBER_OF_PRESET_FIELDS 18
#define MAX_KEY_BUF_SZ   ((SAIDMGR_KEYSIZE+1) * 2)
#define BYTE_SIZE 8

extern int tracesaidmgrp; int tracesaidmgrp;

static XObj             saidmgrOpen( XObj, XObj, XObj, Part * );
static xkern_return_t   saidmgrOpenDone( XObj, XObj, XObj, XObj );
static xkern_return_t   saidmgrClose( XObj );
static int              saidmgrControl( XObj, int, char *, int );
static xkern_return_t   saidmgrDemux( XObj, XObj, Msg * );

static void		saidmgr_getproc_sessn(XObj);
static long             saidmgrGetSaid( void *, char *, long, void * );
static void 		saidmgrPresetKeys(XObj);
static int 		saidmgrSetKeys(XObj, PSTATE *, XSaid, IPhost, KMA_MAPPING, char *, char *);
static int 		saidmgrInitKeys(XObj, SSTATE *, PSTATE * , XSaid);
static void 		saidmgrRemoveKeys(XObj, PSTATE * , XSaid , IPhost);
static int		keystring2key( char *, char *);
static int		GetMyIPAddress( PSTATE *, IPhost *);

static xkern_return_t romallowalgorithms(XObj, char **, int, int, void *);


static XObjRomOpt rmopts[1] = {{ "allowalgorithms", 2, romallowalgorithms}
				};



static xkern_return_t
romallowalgorithms(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  int i;
  XAlgorithm algo;
  Part p[2];
  PSTATE     *ps = (PSTATE *)(self->state);

  for(i=2; i<fields; i++)
  {
    xTraceP1(self, TR_EVENTS, "try algorithm %s", ptr[i]);
    algo = relProtNumById(protTblGetId(ptr[i]),self);
    if ( algo != 0)
    {
	int x;
        for(x=0; ps->allowed_allgorithm[x] != 0; x++);
	if ( x >= SAIDMGR_MAX_ALLOWED_ALGORITHM) return XK_FAILURE;
	ps->allowed_allgorithm[x] =  algo;
    	xTraceP1(self, TR_EVENTS, "allow algorithm %u", algo);
    	partInit(p,2);
    	partPush(p[REMOTE_PART], ANY_HOST, 0);
   	partPush(p[LOCAL_PART], (char *) &algo, sizeof(XAlgorithm));
   	xOpenEnable(self,self, ps->kx, p); 
    }
  }
  return XK_SUCCESS;
}


void saidmgr_init( self )
    XObj self;
{
    int         i;
    int 	j=0;
    PSTATE      *ps;
    XTime	t;

    xAssert(xIsProtocol(self));
    ps = X_NEW(PSTATE);
    self->state = (VOID *)ps;


    if (self->numdown < 3) {
      xTraceP1(self, TR_ERRORS,
	       "init: Illegal lower level protocol configuartion %d", 
	       self->numdown);
      return ;
    }

    xGetTime(&t); /* to get SPIs */
    srand(t.sec);

    ps->spi_map = mapCreate(SAIDMGR_SPI_MAP_SIZE,sizeof(XSaid));
    ps->km_list = (XObj *)xMalloc((self->numdown-2)*sizeof(XObj));
    ps->km_name = (int *)xMalloc((self->numdown-2)*sizeof(int));
    ps->said	= 4711;
    bzero((char *)ps->allowed_allgorithm,
	sizeof(ps->allowed_allgorithm[0]) * SAIDMGR_MAX_ALLOWED_ALGORITHM);
    semInit(&ps->inc_mutex, 1);
	
    /* protocol 0 is key exchange, protocol 1+2 is key manager*/

    ps->kx = xGetDown(self, j++);
    ps->vnet = xGetDown(self, j++);
    ps->kma = xGetDown(self, j++);
    ps->kmb = xGetDown(self, j++);
    ps->km_algo_db = xGetDown(self, j++);
    for ( i=0; i < self->numdown-5; i++ ) {
        XObj    llp;
	char * buf;
	int x;

        llp = xGetDown(self, j++);
        ps->km_list[i] = llp;
        if ( ! xIsProtocol(llp) ) {
            xTraceP1(self, TR_ERRORS, "can't find protocol %d", i);
            xError("SAIDMGR init errors -- not initializing");
            return;
        } else {
            xTraceP2(self, TR_GROSS_EVENTS, "llp %d is %s", i, llp->name);
        }
	buf = xMalloc(strlen(llp->instName)+1);
	strcpy(buf,llp->instName);
	for(x=0; (buf[x] != 0) && (buf[x] != '_'); x++);
	buf[x]=0;
	if ((ps->km_name[i] =  relProtNumById(protTblGetId(buf),self))<=0) 
	{
            xTraceP1(self, TR_ERRORS, "wrong instance name %s", llp->fullName);
            xError("SAIDMGR init errors -- not initializing");
	    xFree(buf);
            return;
	}
	xTraceP3(self, TR_GROSS_EVENTS,"Set key manager table entry %u %u %x ",
		i,ps->km_name[i],ps->km_list[i]);
	xFree(buf);
    }
    ps->km_name[i] = 0;

    findXObjRomOpts(self, rmopts, 1, 0);

    self->control = saidmgrControl;
    self->open = saidmgrOpen;
    self->demux = saidmgrDemux;
    self->opendone = saidmgrOpenDone;
    saidmgrPresetKeys(self);
}

static void
saidmgr_getproc_sessn(s)
    XObj s;
{
  xAssert(xIsSession(s));
  s->control = saidmgrControl;
  s->close = saidmgrClose;
}

static long saidmgrGetSaid(dst, src, count, arg)
     void *dst;
     char *src;
     long count;
     void *arg;
{
	unsigned long *	ds = (unsigned long *) dst;
	unsigned long  d =0;
	memcpy((((char *)&d)+(4-count)),src,count);
	*ds = ntohl (d);
	return count;
 }

static int GetMyIPAddress( PSTATE * p, IPhost * host)
{
	int number_of_hosts;
	IPhost * host_list;
	xControl(p->vnet,GETMYHOSTCOUNT,(char *) &number_of_hosts,sizeof(int));
	if (number_of_hosts <= 0 )
		return -1;
	host_list = (IPhost *) xMalloc(sizeof(IPhost) * number_of_hosts);
	xControl(p->vnet,GETMYHOST,(char *) host_list,sizeof(IPhost)*number_of_hosts);
	*host = host_list[0];
	xFree((char *)host_list);
	return 1;
}

static int keystring2key( char * keystring, char * key)
{
	unsigned int x;
	unsigned int sz;
	sz = strlen(keystring);
	if ((sz % 2) == 1)  {
		keystring[sz]='0';
		sz ++;
		keystring[sz]=0;
	}
	sz/=2;
	if (sz < 8) {xError("SAIDMGR: preset key to small"); return -1;}
	if (sz > SAIDMGR_KEYSIZE) {xError("SAIDMGR: preset key to long"); return -1;}
	for(x=0; x<sz; x++)
	{
		char buf[3];
		int i;
		buf[2]=0;
		buf[1]=keystring[x*2+1];
		buf[0]=keystring[x*2];
		sscanf(buf,"%x",&i);
		key[x]=(char) i;
	}

	return 1;
}

static void  saidmgrPresetKeys(XObj self)
{	
	FILE * keyfile;
 	KMA_MAPPING kmma;
	char local_key[MAX_PRESET_KEYS];
	char remote_key[MAX_PRESET_KEYS];
	XSaid lsaid;
	IPhost myhost;
	char local_algorithm_buf[MAX_KEY_BUF_SZ];
	char local_key_buf[MAX_KEY_BUF_SZ];
	char remote_algorithm_buf[MAX_KEY_BUF_SZ];
	char remote_key_buf[MAX_KEY_BUF_SZ];
	PSTATE	*ps = (PSTATE *) self->state;
	int la,lb,lc,ld,ra,rb,rc,rd;
	int matched;
	int lineno = 0;

	bzero(local_key,MAX_PRESET_KEYS);
	bzero(remote_key,MAX_PRESET_KEYS-1);

	if ((keyfile = fopen("./preset_keys","r")) == NULL) 
		return;
	while ((matched = fscanf(keyfile,"%u;%d.%d.%d.%d;%[^;];%u;%u;%[^=]=%u;%u.%u.%u.%u;%[^;];%u;%u;%[^\n]\n",
		&lsaid,&la,&lb,&lc,&ld,local_algorithm_buf,
		&kmma.local_parameter, &kmma.local_mode, local_key_buf,
		&kmma.remote_said, &ra, &rb, 
		&rc, &rd, remote_algorithm_buf,
		&kmma.remote_parameter, &kmma.remote_mode,
		remote_key_buf))
			 != EOF)
	{
		int local_address_ok = 1;

		if ((la != 0) || (lb != 0) || (lc != 0) || (ld != 0))  
			{ myhost.a=la; myhost.b=lb; myhost.c=lc; myhost.d=ld; }
		else
			local_address_ok = GetMyIPAddress(ps, &myhost);  

		kmma.remote_host.a=ra; kmma.remote_host.b=rb;
		kmma.remote_host.c=rc; kmma.remote_host.d=rd;

    		xIfTrace (saidmgrp, TR_MORE_EVENTS)
		{
		        char	buf[512];
			
		        sprintf(buf,
				"preset keys %u matched"
				"%u;%u.%u.%u.%u;%s;%u;%u;%s="
				"%u;%u.%u.%u.%u;%s;%u;%u;%s",
				matched, lsaid,
				myhost.a,myhost.b,myhost.c,myhost.d,
				local_algorithm_buf,
				kmma.local_parameter, kmma.local_mode,
				local_key_buf,
				kmma.remote_said,
				kmma.remote_host.a, kmma.remote_host.b, 
				kmma.remote_host.c, kmma.remote_host.d,
				remote_algorithm_buf,
				kmma.remote_parameter, kmma.remote_mode,
				remote_key_buf);
			xTraceP1(self, TR_MORE_EVENTS, "%s", buf);
 		}

		if ((matched == NUMBER_OF_PRESET_FIELDS)
		   && ((kmma.local_algorithm = relProtNumById(protTblGetId(local_algorithm_buf),self)) > 0)
		   && ((kmma.remote_algorithm = relProtNumById(protTblGetId(remote_algorithm_buf),self)) > 0)
		   && (keystring2key(local_key_buf, local_key) > 0)
		   && (keystring2key(remote_key_buf, remote_key) > 0)
		   && (local_address_ok > 0) )
		{
 		        if (saidmgrSetKeys(self, ps, lsaid, myhost, kmma, local_key, remote_key) <0)
          			xError("SAIDMGR couldn't set entry in preset_key file");
 		}
		else {
		  char c;
           		xError("SAIDMGR illegal algorithm entry in preset_key file");
			if (matched < NUMBER_OF_PRESET_FIELDS) {
			  while (((c=fgetc(keyfile)) != EOF) && (c != '\n'));
			  xTraceP1(self, TR_SOFT_ERRORS,
				   "too few items in line %d\n", lineno);
			}
		      }
		lineno++;
 	}
	fclose(keyfile);
}

static int saidmgrSetKeys(XObj self, PSTATE * ps, XSaid said, IPhost myhost,  KMA_MAPPING kmma,
			 char* lkey, char * rkey)
{
     XObj	kms;
     Part	kmp[1];	
     int 	i;
     unsigned char      tmp[3];
     unsigned char 	key;

     /* setup iv and hash_length */
     partInit(kmp,1);
     key = (unsigned char) kmma.local_algorithm;
     partPush(kmp[0],(char *)&key,sizeof(key));
     kms = xOpen(self,self,ps->km_algo_db, kmp);	
     if (kms == ERR_XOBJ) 
     {
 	kmma.local_algorithm = 0;
	kmma.local_iv_length = 0;
	kmma.local_hash_size = 0;
	kmma.local_padding_type = SA_NO_PADDING;
     }
     else {
	    if ( xControl(kms, KM_RESOLVE, (char*) tmp, 4*sizeof(int)) > 0) 
            {
		kmma.local_algorithm = (int) tmp[0];
		kmma.local_iv_length = (int) tmp[1];
		kmma.local_hash_size = (int) tmp[2];
		kmma.local_padding_type = (int) tmp[3];
	    }
	 else
	   {
	     kmma.local_algorithm = 0;
	     kmma.local_iv_length = 0;
	     kmma.local_hash_size = 0;
	     kmma.local_padding_type = SA_NO_PADDING;
	   }
      
     	xClose(kms); 
     }

     partInit(kmp,1);
     key = (unsigned char) kmma.remote_algorithm;
     partPush(kmp[0],(char *)&key,sizeof(key));
     kms = xOpen(self,self,ps->km_algo_db, kmp);	
     if (kms == ERR_XOBJ) 
     {
 	kmma.remote_algorithm = 0;
	kmma.remote_iv_length = 0;
	kmma.remote_hash_size = 0;
	kmma.remote_padding_type = SA_NO_PADDING;
     }
     else 
     {
	    if ( xControl(kms, KM_RESOLVE, (char*) tmp, 4*sizeof(int)) > 0) 
	    {
		kmma.remote_algorithm = (int) tmp[0];
		kmma.remote_iv_length = (int) tmp[1];
		kmma.remote_hash_size = (int) tmp[2];
		kmma.remote_padding_type = (int) tmp[3];
	    }
	 else
	 {
	     kmma.remote_algorithm = 0;
	     kmma.remote_iv_length = 0;
	     kmma.remote_hash_size = 0;
	     kmma.remote_padding_type = SA_NO_PADDING;
	 }
      
     	xClose(kms); 
      }


     partInit(kmp,1);
     partPush(kmp[0],(char *)&said,sizeof(XSaid));	
     partPush(kmp[0],(char *)&myhost,sizeof(IPhost));	
     kms = xOpen(self, self, ps->kma , kmp);
     if (kms == ERR_XOBJ) 
     {
        xTraceP0(self, TR_ERRORS, "can't open KM/a");
	return -1;
     }
     xControl(kms, KM_SET, (char*) &kmma, sizeof(KMA_MAPPING));
     xClose(kms); 
     xTraceP4(self, TR_DETAILED,
	      "got from algorithm_db local prot %u ivsz %u hashsz %u paddingType %u",
		kmma.local_algorithm,kmma.local_iv_length,
		kmma.local_hash_size,kmma.local_padding_type);
     xTraceP4(self, TR_DETAILED,
	      "got from algorithm_db remote prot %u ivsz %u hashsz %u paddingType %u",
		kmma.remote_algorithm, kmma.remote_iv_length,
		kmma.remote_hash_size,kmma.remote_padding_type);
     for(i = 0; ps->km_name[i] > 0 ; i++)
     if(ps->km_name[i] == kmma.local_algorithm) 
     {
	int key_sz;
     	partInit(kmp,1);
     	partPush(kmp[0],(char *)&said,sizeof(XSaid));	
     	partPush(kmp[0],(char *)&myhost,sizeof(IPhost));	
	xTraceP3(self, TR_DETAILED,"Set key in km %u %u %x ",
		i,ps->km_name[i],ps->km_list[i]);
     	kms = xOpen(self, self, ps->km_list[i] , kmp);
        if (kms == ERR_XOBJ) 
	{	
		saidmgrRemoveKeys(self,ps,said,myhost);
          	xTraceP1(self, TR_ERRORS, "can't open KM for algorithm %u",
			kmma.local_algorithm );
		return -1;
	}
     	xControl(kms, KM_KEYSIZE, (char*) &key_sz, sizeof(int));
	if (key_sz>SAIDMGR_KEYSIZE)
	{	/* dheckx doesn't have a greater secret */
     		xClose(kms);
		saidmgrRemoveKeys(self,ps,said,myhost);
           	xTraceP1(self, TR_ERRORS, "needs to large secret %u",
			 kmma.local_algorithm );
 		return -1;
	}
     	xControl(kms, KM_SET, lkey, key_sz);
     	xClose(kms);
     } 

     for(i = 0; ps->km_name[i] > 0 ; i++)
     if(ps->km_name[i] == kmma.remote_algorithm) 
     {
	int key_sz;
     	partInit(kmp,1);
     	partPush(kmp[0],(char *)&kmma.remote_said,sizeof(XSaid));	
     	partPush(kmp[0],(char *)&kmma.remote_host,sizeof(IPhost));	
	xTraceP3(self, TR_DETAILED,"Set key in km %u %u %x ",
		i,ps->km_name[i],ps->km_list[i]);
     	kms = xOpen(self, self, ps->km_list[i] , kmp);
        if (kms == ERR_XOBJ) 
	{	
		saidmgrRemoveKeys(self,ps,said,myhost);
          	xTraceP1(self, TR_ERRORS, "can't open KM for algorithm %u",
			kmma.remote_algorithm );
		return -1;
	}
     	xControl(kms, KM_KEYSIZE, (char*) &key_sz, sizeof(int));
	if (key_sz>SAIDMGR_KEYSIZE)
	{	/* dheckx doesn't have a greater secret */
     		xClose(kms);
		saidmgrRemoveKeys(self,ps,said,myhost);
           	xTraceP1(self, TR_ERRORS, "km needs to large secret %u",
			 kmma.remote_algorithm );
 		return -1;
	}
     	xControl(kms, KM_SET, rkey, key_sz);
     	xClose(kms);
     } 
     /* add to spi_map */
     {
	SPILIST ** p;
	SPILIST * c;
	c = (SPILIST *) xMalloc(sizeof(SPILIST));
	c->local_spi = said;
	c->remote_spi = kmma.remote_said;
	if (XK_SUCCESS == mapResolve(ps->spi_map,&kmma.remote_host,&p))
	{
		SPILIST * e = *p;
		while ( (e->next!=NULL) && (e->local_spi != c->local_spi) )
			e = e->next;
		if (e->local_spi != c->local_spi)
		{			
			c->next = *p;
			*p = c;
		}
		else
		{
			xFree((char *) c);
		}	
	}
	else
	{
		SPILIST ** n;
		c->next = NULL;
		n = (SPILIST **) xMalloc(sizeof(SPILIST *));
		*n = c;
		mapBind(ps->spi_map,&kmma.remote_host,n);
	}
     } 
     return 1;
}


static int saidmgrInitKeys(XObj self, SSTATE * ss, PSTATE * ps, XSaid said)
{
     XObj	kms;
     Part	kmp[1];	
     KMA_MAPPING kmma;	
     KMB_MAPPING kmmb;
     int	 ret;	

     xTraceP0(self, TR_EVENTS, "Init Keys");
     partInit(kmp,1);
     partPush(kmp[0],(char *)&said,sizeof(XSaid));	
     kms = xOpen(self, self, ps->kmb , kmp);
     if (kms == ERR_XOBJ)
     {
        xTraceP0(self, TR_ERRORS, "can't open KM/b");
	return -1;
     }
     xControl(kms, KM_RESOLVE, (char*) &kmmb, sizeof(KMB_MAPPING));
     xControl(kms, KM_INVALIDATE, NULL, 0);
     xClose(kms);  
     
     if (ss != NULL)
     {
     	ss->myhost=kmmb.local_host;
	ss->remotehost=kmmb.remote_host;
     }

     /* timeouts should be transfered to timeout manager */
     kmma.local_algorithm = kmmb.local_algorithm;
     kmma.local_parameter = kmmb.local_parameter;
     kmma.local_mode = kmmb.local_mode;
     kmma.remote_said = kmmb.remote_said;
     kmma.remote_host = kmmb.remote_host;
     kmma.remote_algorithm = kmmb.remote_algorithm;
     kmma.remote_parameter = kmmb.remote_parameter;
     kmma.remote_mode = kmmb.remote_mode;

     xIfTrace(saidmgrp, TR_DETAILED)
     {
        int x;
	xLogP0(self, TR_DETAILED, "----------------------\n Enter SA :\n");
	xLogP4(self, TR_DETAILED, "l_address %u.%u.%u.%u\n",
	       kmmb.local_host.a,kmmb.local_host.b,
	       kmmb.local_host.c,kmmb.local_host.d);
	xLogP4(self, TR_DETAILED,
	       "l_spi = %u; l_algorithm = %u; l_parameter = %u; l_mode = %u\n",
	       said,
	       kmma.local_algorithm, kmma.local_parameter, kmma.local_mode);
	
	xLogP0(self, TR_DETAILED, "local key = ");
	for(x=0; x < SAIDMGR_KEYSIZE; x++)
	  xPrintf("%02x ", (unsigned char) kmmb.local_key[x]);
	xPrintf("\n");
	
	xLogP4(self, TR_DETAILED, "r_address %u.%u.%u.%u\n",
	       kmmb.remote_host.a,kmmb.remote_host.b,
	       kmmb.remote_host.c,kmmb.remote_host.d);
 	xLogP4(self, TR_DETAILED,
	       "r_spi = %u; r_algorithm = %u; r_parameter = %u; r_mode = %u\n",
	       kmma.remote_said,
	       kmma.remote_algorithm, kmma.remote_parameter, kmma.remote_mode);
	
	xLogP0(self, TR_DETAILED, "remote key = ");
	for(x=0; x < SAIDMGR_KEYSIZE; x++)
		xPrintf("%02x ",(unsigned char) kmmb.remote_key[x]);
	xPrintf("\n");
	
	xLogP0(self, TR_DETAILED, "----------------------\n");
     }

     ret = saidmgrSetKeys(self, ps, said, kmmb.local_host, kmma, kmmb.local_key, kmmb.remote_key);
     /* have to zero out keys */
     xTraceP1(self, TR_EVENTS, "Init Keys finished %08x",ret);
     return ret;
}

static void saidmgrRemoveKeys(XObj self,PSTATE * ps, XSaid said, IPhost myhost)
{
     XObj	kms;
     Part	kmp[1];	
     int 	i;
     KMB_MAPPING kmmb;


     partInit(kmp,1);
     partPush(kmp[0],(char *)&said,sizeof(XSaid));	
     partPush(kmp[0],(char *)&myhost,sizeof(IPhost));	
     kms = xOpen(self, self, ps->kma , kmp);
     if (kms != ERR_XOBJ)
     {
     	xControl(kms, KM_RESOLVE, (char*) &kmmb, sizeof(kmmb));
     	xControl(kms, KM_INVALIDATE, NULL, 0);
     	xClose(kms);
     } 
     
     for(i = 0; ps->km_name[i] > 0 ; i++)
     if(ps->km_name[i] == kmmb.local_algorithm) 
     {
     	partInit(kmp,1);
     	partPush(kmp[0],(char *)&said,sizeof(XSaid));	
     	partPush(kmp[0],(char *)&myhost,sizeof(IPhost));	
    	kms = xOpen(self, self, ps->km_list[i] , kmp);
        if (kms != ERR_XOBJ)
	{
     	    xControl(kms, KM_INVALIDATE, NULL, 0);
     	    xClose(kms);
	} 
     }
     for(i = 0; ps->km_name[i] > 0 ; i++)
     if(ps->km_name[i] == kmmb.remote_algorithm) 
     {
     	partInit(kmp,1);
     	partPush(kmp[0],(char *)&kmmb.remote_said,sizeof(XSaid));	
      	partPush(kmp[0],(char *)&kmmb.remote_host,sizeof(IPhost));	
   	kms = xOpen(self, self, ps->km_list[i] , kmp);
        if (kms != ERR_XOBJ)
	{
     	    xControl(kms, KM_INVALIDATE, NULL, 0);
     	    xClose(kms);
	} 
     }
     /* remove from spi_map */
     {
	SPILIST ** p;
	SPILIST * c;
	if (XK_SUCCESS == mapResolve(ps->spi_map,&kmmb.remote_host,&p))
	{
		c = *p;	
		if  (c->local_spi == said)
		{
		    	if (c->next == NULL)
			{ 
				xFree((char*) p);
				xFree((char*)c);
				mapUnbind(ps->spi_map,&kmmb.remote_host);
			}
			else
			{
				*p = c->next;
				xFree((char *) c);
			}
		}
		else
		{
			SPILIST * e;
			e = c;
			c = c->next;
			while ((c->next != NULL) && (c->local_spi != said))
				e=c; c=c->next;
			if ( c->local_spi == said )
			{
				e->next = c->next;
				xFree((char *)c);
			}	
		}
	}
     } 
}

static XObj
saidmgrOpen( self, hlpRcv, hlpType, p )
    XObj        self, hlpRcv, hlpType;
    Part        *p;
{
    
     PSTATE     *ps = (PSTATE *)(self->state);
     SSTATE	*ss;
     XObj	lls;
     XObj	s;
     int	x;
     void 	*tmp1;
     void	*tmp2;
     IPhost 	localhost;

     xTraceP0(self, TR_EVENTS, "open");
	
     s = xCreateSessn(saidmgr_getproc_sessn ,hlpRcv, hlpType, self,0, NULL);  
     ss = (SSTATE *) xMalloc(sizeof(SSTATE));
     s->state = (SSTATE *) ss; 
     semWait(&ps->inc_mutex);   
     ss->lsaid = rand();
     ss->lsaid = (ss->lsaid < 2000) ? ss->lsaid + 2000 : ss->lsaid; 
	/* first 2000 SPIs for preset keys*/
     semSignal(&ps->inc_mutex);
     xTraceP1(self, TR_EVENTS, "assigned lsaid %u", ss->lsaid);
     if (partLen(p)<SPI_PART)
     {
          xError("SAIDMGR algorithm and ipsec mode have to be specified to open SAIDMGR session.");
 	  return ERR_XOBJ;
     }
     tmp1=partPop(p[SPI_PART]);
     tmp2=partPop(p[SPI_PART]);
     if (tmp2==NULL)
     {
          xError("SAIDMGR algorithm and ipsec mode have to be specified to open SAIDMGR session.");
 	  return ERR_XOBJ;
     }
     ss->algorithm = *((XAlgorithm *)tmp1);
     ss->mode = *((XIpsecMode *)tmp2);
     for(x=0; (ps->allowed_allgorithm[x] != 0) && 
	(ps->allowed_allgorithm[x] != ss->algorithm) ; x++);
     if (ps->allowed_allgorithm[x] == 0) 
     {
        xError("SAIDMGR requested algorithm rejected by host policy.");
	return ERR_XOBJ;
     }
     xTraceP1(self, TR_EVENTS, "use algorithm %u", ss->algorithm);

     tmp1=partPop(p[LOCAL_PART]);
     if ((tmp1 == NULL) || (tmp1 == ANY_HOST))
     {
	if ( GetMyIPAddress(ps, &localhost) < 0 )
     	{
            xError("SAIDMGR couldn't retrieve local host address.");
 	    return ERR_XOBJ;
     	}
	partPush(p[LOCAL_PART],(char *)&localhost,sizeof(IPhost));	
     }	
     else
	partPush(p[LOCAL_PART],(char *)tmp1,sizeof(IPhost));	


     partPush(p[SPI_PART],(char *)&ss->mode,sizeof(XIpsecMode));	
     partPush(p[SPI_PART],(char *)&ss->algorithm,sizeof(XAlgorithm));	
     partPush(p[SPI_PART],(char *)&ss->lsaid,sizeof(XSaid));	

     lls = xOpen(self, self, ps->kx , p);
     if (lls == ERR_XOBJ) return ERR_XOBJ;
     xClose(lls);
     if (saidmgrInitKeys(self, ss, ps, ss->lsaid)<0) return ERR_XOBJ;
     xTraceP0(self, TR_EVENTS, "open done");

     return s;
}

static xkern_return_t
saidmgrOpenDone( self, lls, llp, hlpType)
     XObj self, lls, llp, hlpType;
{

  PSTATE     *ps = (PSTATE *)(self->state);
  XSaid lsaid;

     
  semWait(&ps->inc_mutex);   
  lsaid = rand();
  lsaid = (lsaid < 2000) ? lsaid + 2000 : lsaid; 
	/* first 2000 SPIs for preset keys*/
  semSignal(&ps->inc_mutex);

  return lsaid;
}

static xkern_return_t
saidmgrDemux(self, transport_s, dg)
    XObj self;
    XObj transport_s;
    Msg *dg;
{

  PSTATE     *ps = (PSTATE *)(self->state);
  XSaid lsaid;
  
  xTraceP0(self, TR_EVENTS, "demux");
  msgPop(dg, saidmgrGetSaid, &lsaid, sizeof(XSaid),NULL);
  if (saidmgrInitKeys(self,NULL, ps, lsaid)<0) return XK_FAILURE;
  xTraceP0(self, TR_EVENTS, "demux end");
   
  return XK_SUCCESS; 
}

static xkern_return_t
saidmgrClose(s)
     XObj s;
{
  SSTATE     *ss = (SSTATE *)(s->state);
  if (s->rcnt == 0)
  {
  	saidmgrRemoveKeys(s->myprotl,(PSTATE *)(s->myprotl->state), ss->lsaid, ss->myhost);
  	xFree((char *)(s->state));
  }
  return XK_SUCCESS;
}

static int
saidmgrControl(self, opcode, buf, len)
    XObj self;
    int opcode;
    char *buf;
    int len;
{
    
    xAssert(xIsProtocol(self)||xIsSession(self));

    xTraceP1(self, TR_DETAILED,"xControl called %u", opcode);

    switch (opcode) {
/* The following two options are unecessary and will be removed in future releases (os)*/
      case SAIDMGR_GETREMOTEALGORITHM:
        {
         return XK_FAILURE;
        }
      case SAIDMGR_GETLSAID:
        { 
		if (xIsSession(self))
		{	
			return ((SSTATE *)self->state)->lsaid;
		}
		else
			return -1;
        }

      case SAIDMGR_RESOLVE_NAME:
        { 
		if (xIsSession(self) || xIsProtocol(self) )
			return relProtNumById(protTblGetId(buf),self);
		else
			return -1;
        }


      case SAIDMGR_RESCAN_PRESET_KEY_FILE :
	{
		if (xIsProtocol(self))
		{
	   	    	saidmgrPresetKeys(self);	
			return XK_SUCCESS;
		}
		else
			return XK_FAILURE;
	}
      case SAIDMGR_GET_SPIS_FOR_HOST :
	{
		SPILIST ** p;
		if (xIsProtocol(self) && (len >= sizeof(IPhost))
			&& (XK_SUCCESS == mapResolve(((PSTATE*)self->state)->spi_map,buf,&p)) )
		{
			int x = 0;
			XSaid * ar = (XSaid *) buf;
			SPILIST * c = *p;
			while ( (c!=NULL) && (((x+2) * sizeof(XSaid)) <= len) )
			{
				ar[x] = c->local_spi;
				x++;
				ar[x] = c->remote_spi;
				x++;	
				c = c->next;			
			}	
			return x/2;
		}
		else
			return -1;
	}
      case GETPEERHOST :
        { 
		if (xIsSession(self))
		{	
	  		checkLen(len, sizeof(IPhost));
			memcpy(buf,(char *)&((SSTATE *)self->state)->remotehost,sizeof(IPhost));
			return XK_SUCCESS;
		}
		else
			return XK_FAILURE;
	}

      case GETMYHOST :
	{
		if (xIsSession(self))
		{	
	  		checkLen(len, sizeof(IPhost));
			memcpy(buf,(char *)&((SSTATE *)self->state)->myhost,sizeof(IPhost));
			return XK_SUCCESS;
		}
		else
			return XK_FAILURE;
	}      

      default:
	return XK_FAILURE;
    }
}





