/*
 * $RCSfile: vsecselspi.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * $Log: vsecselspi.c,v $
 * Revision 1.10  1996/02/15 21:58:26  mjk
 * added conditional #undef min before #define
 *
 *
 */

/*
 *   This is a security selection protocol
 *
 */

#include "xkernel.h"
#include "km.h"
#include "saidmgr.h"
#include "vsecselspi.h"
#include "vsecselspi_i.h"


int tracevsecselspip;

#ifdef __STDC__

static XObj		vsecselspiOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	vsecselspiOpenDone( XObj, XObj, XObj, XObj );
static xkern_return_t	vsecselspiOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t	vsecselspiOpenEnable( XObj, XObj, XObj, Part * );
static int		vsecselControlProtl( XObj, int, char *, int );

static xkern_return_t	xCopyOpenDisable( XObj, XObj, XObj, Part *, XAlgorithm );
static xkern_return_t	xCopyOpenEnable( XObj, XObj, XObj, Part *, XAlgorithm );
#endif


#define xGetParticipants(lls, p) \
  xControl(lls, GETPARTICIPANTS, (char *)p, sizeof(p))

#ifdef min
#undef min
#endif
#define min(a,b) (a)<(b)?(a):(b)

#define COPYBUFLEN 1000

static int
vsecselControlProtl( self, opcode, buf, len )
    XObj self;
    int opcode, len;
    char *buf;
{
   char bufcopy[MAX_CONTROL_BUF];
   PState *ps = (PState *)(self->state);
   int i;
   int ret = -1;

   for(i=0; (ps->llptypes[i] != 0); i++) 
    		ret = xControl(ps->llplist[i], opcode, bufcopy,min(len, sizeof(bufcopy)));
   if (ps->ip != NULL)
		ret = xControl(ps->ip, opcode, bufcopy, min(len, sizeof(bufcopy)));

   bcopy(buf, bufcopy, min(len, sizeof(bufcopy)));
 
   /* only information from the last protocol can be returned */
   return ret;
}

static xkern_return_t	xCopyOpenDisable( XObj self, XObj type , XObj llp, Part * p, XAlgorithm algo )
{
    char buf[COPYBUFLEN];
    int	 sz=COPYBUFLEN;
    Part p2[3];
    if (partLen(p) == 3)
    {
    	partExternalize(p,buf,&sz);
   	partInternalize(p2,buf);
    }
    else
    {
	Part *p3;
	int	x;
	int	len;
	p3 = (Part *) xMalloc(sizeof(Part)*partLen(p));
    	partExternalize(p,buf,&sz);
   	partInternalize(p3,buf);
	partInit(p2,3);
	for(x=0; x < partLen(p3); x++)
	{
		len = partStackTopByteLen(p3[x]);
		partPush(p2[x],partPop(p3[x]),len);
	} 
	xFree((char *) p3);
    }
    partPush(p2[2], &algo, sizeof(XAlgorithm));
    return xOpenDisable(self,type,llp,p2);
}

static xkern_return_t	xCopyOpenEnable( XObj self, XObj type, XObj llp, Part * p, XAlgorithm algo )
{
    char buf[COPYBUFLEN];
    int	 sz=COPYBUFLEN;
    Part p2[3];
    if (partLen(p) == 3)
    {
    	partExternalize(p,buf,&sz);
   	partInternalize(p2,buf);
    }
    else
    {
	Part *p3;
	int	x;
	int	len;
	p3 = (Part *) xMalloc(sizeof(Part)*partLen(p));
    	partExternalize(p,buf,&sz);
   	partInternalize(p3,buf);
	partInit(p2,3);
	for(x=0; x < partLen(p3); x++)
	{
		len = partStackTopByteLen(p3[x]);
		partPush(p2[x],partPop(p3[x]),len);
	} 
	xFree((char *) p3);
    }
    partPush(p2[2], &algo, sizeof(XAlgorithm));
    return xOpenEnable(self,type,llp,p2);
  
}



/*
 *  vsecselspiOpen
 *
 *
 */
static XObj
vsecselspiOpen( self, hlpRcv, hlpType, p )
    XObj	self, hlpRcv, hlpType;
    Part 	*p;
{
    Part	kmp[2];
    XObj	lls;
    XSaid	lsaid;
    KMA_MAPPING kmma;	
    XObj	kms;
    int		i;
    PState	*ps = (PState *)(self->state);
    IPhost	myhost;
    
    xTrace0(vsecselspip, TR_MAJOR_EVENTS, "VSECSEL open");


    /* 
      The service type will be based on the local said
     */

     if (partLen(p) < 3)
     {
    	xTrace0(vsecselspip, TR_EVENTS, "VSECSEL use IP directly.");
	lsaid = VSECSEL_SAID_IP;
     }
     else
     {			
     	lsaid = *(XSaid *)partPop(p[2]);
     }
     if (lsaid == VSECSEL_SAID_IP)
     {
	if (ps->ip == NULL)
	{
	    xTrace0(vsecselspip, TR_EVENTS, "No direct IP available.");
	    return ERR_XOBJ;
	}
     	if ((lls = xOpen(hlpRcv, hlpType, ps->ip, p))
	    != ERR_XOBJ )
	  {
	    xTrace0(vsecselspip, TR_MAJOR_EVENTS,
		    "VSECSEL open of llp IP successful ");
	    /* 
	     * This session is passing through the open of more than
	     * one protocol, so its reference count is being
	     * artificially increased.  We'll correct for this level
	     * before returning the session.
	     *
	     * Should we be scheduling an event to do an xClose on this
	     * session? 
	     */
	    lls->rcnt--;
	    return lls;
	  }
 	xTrace0(vsecselspip, TR_MAJOR_EVENTS,
		    "VSECSEL open of llp IP UNsuccessful ");
	return ERR_XOBJ;
     }

     partPush(p[2], (char *)&lsaid, sizeof(XSaid));
     if (partStackTopByteLen(p[1]) != sizeof(IPhost))
     {
 	xTrace0(vsecselspip, TR_MAJOR_EVENTS,
		    "VSECSEL open error: local host is not specified or not an IPhost");
	return ERR_XOBJ;
     }
     myhost=*(IPhost *)partPop(p[1]);
     partPush(p[1], (char *)&myhost, sizeof(IPhost));

     xTrace1(vsecselspip, TR_MAJOR_EVENTS, "VSECSEL open LSAID %u",lsaid);

     partInit(kmp,1);
     partPush(kmp[0],(char *)&lsaid,sizeof(XSaid));	
     partPush(kmp[0],(char *)&myhost,sizeof(IPhost));	
     kms = xOpen(self, self, ps->km , kmp);
     if (kms == ERR_XOBJ) return ERR_XOBJ;
     xControl(kms, KM_RESOLVE, (char *) &kmma, sizeof(KMA_MAPPING));
     xClose(kms); 

     for(i=0; (ps->llptypes[i] != 0) && (ps->llptypes[i] != kmma.local_algorithm); i++);
 
     if ((ps->llptypes[i] != 0) &&
	 ((lls = xOpen(hlpRcv, hlpType, ps->llplist[i], p))
	    != ERR_XOBJ ))
	  {
	    xTrace1(vsecselspip, TR_MAJOR_EVENTS,
		    "VSECSEL open of llp %d successful", kmma.local_algorithm);
	    /* 
	     * This session is passing through the open of more than
	     * one protocol, so its reference count is being
	     * artificially increased.  We'll correct for this level
	     * before returning the session.
	     *
	     * Should we be scheduling an event to do an xClose on this
	     * session? 
	     */
	    lls->rcnt--;
	    return lls;
	  }
    xTrace0(vsecselspip, TR_MAJOR_EVENTS, "VSECSEL open fails");
    return ERR_XOBJ;
}

/*
 *  vsecselspiOpenEnable
 *
 *    assume there is only one participant, the convention is that it specifies
 *    the local part ... e.g. the local server port for this machine.
 *
 */
static xkern_return_t
vsecselspiOpenEnable( self, hlpRcv, hlpType, p )
    XObj	self, hlpRcv, hlpType;
    Part	*p;
{
    PState	*ps = (PState *)self->state;
    int		i;
    XAlgorithm	algo;

    xTrace0(vsecselspip, TR_MAJOR_EVENTS, "VSECSEL openenable");
    if (partLen(p) < 3) 
    {
    	defaultOpenEnable(ps->passiveMap, hlpRcv, hlpType, (void *)&hlpType);
     	for(i=0; (ps->llptypes[i] != 0); i++) 
    		xCopyOpenEnable(self, hlpType, ps->llplist[i], p, ps->llptypes[i]);
	if (ps->ip != NULL)
		xOpenEnable(self,hlpType, ps->ip, p);
        return XK_SUCCESS;
    } 

    defaultOpenEnable(ps->passiveMap, hlpRcv, hlpType, (void *)&hlpType);

    algo =  *((int *)partPop(p[2]));
 

    xTrace0(vsecselspip, TR_MAJOR_EVENTS, "VSECSEL openenable lower level protocols");
    if (algo == VSECSEL_ANY_ALGORITHM)
    {
    	for(i=0; (ps->llptypes[i] != 0); i++) 
    		xCopyOpenEnable(self, hlpType, ps->llplist[i], p, ps->llptypes[i]);
	if (ps->ip != NULL)
		xOpenEnable(self,hlpType, ps->ip, p);
    } 
    else if (algo == VSECSEL_SAID_IP)
    {
	if (ps->ip != NULL)
		xOpenEnable(self,hlpType, ps->ip, p);
	else
		return XK_FAILURE;
    }
    else
    {
    	for(i=0; (ps->llptypes[i] != 0) && (ps->llptypes[i] != algo); i++); 
    	if (ps->llptypes[i] != 0)
    		return xCopyOpenEnable(self, hlpType, ps->llplist[i], p,algo);
    }
    return XK_SUCCESS;
}

static xkern_return_t
vsecselspiOpenDisable( self, hlpRcv, hlpType, p )
    XObj self, hlpRcv, hlpType;
    Part *p;
{
    PState	*ps = (PState *)self->state;
    int		i;
    XAlgorithm	algo;

    xTrace0(vsecselspip, TR_MAJOR_EVENTS, "VSECSEL open disable");

    if (partLen(p) < 3) 
    {
    	defaultOpenDisable(ps->passiveMap, hlpRcv, hlpType, p);
    	for(i=0; (ps->llptypes[i] != 0); i++) 
		xCopyOpenDisable(self, hlpType, ps->llplist[i], p, ps->llptypes[i]);
	if (ps->ip != NULL)
		xOpenDisable(self,hlpType, ps->ip, p);
        return XK_SUCCESS;
    } 

    defaultOpenDisable(ps->passiveMap, hlpRcv, hlpType, p);

    algo =  *((XAlgorithm *)partPop(p[2]));
 
    if (algo == VSECSEL_ANY_ALGORITHM)
    {
    	for(i=0; (ps->llptypes[i] != 0); i++) 
		xCopyOpenDisable(self, hlpType, ps->llplist[i], p, ps->llptypes[i]);
	if (ps->ip != NULL)
		xOpenDisable(self,hlpType, ps->ip, p);
    } 
    else if (algo == VSECSEL_SAID_IP)
    {
	if (ps->ip != NULL)
		xOpenDisable(self,hlpType, ps->ip, p);
	else
		return XK_FAILURE;
    }
    else
    {
    	for(i=0; (ps->llptypes[i] != 0) && (ps->llptypes[i] != algo); i++); 
    		if (ps->llptypes[i] != 0)
			return xCopyOpenDisable(self, hlpType, ps->llplist[i], p, algo);
    }
    return XK_SUCCESS;
}

/*
 *  vsecselspiOpenDone
 *
 */
static xkern_return_t
vsecselspiOpenDone( self, lls, llp, hlpType )
    XObj self, lls, llp, hlpType;
{
    PState	*ps = (PState *)self->state;
    Enable	*e;
    xkern_return_t	val;
    
    xTrace0(vsecselspip, TR_MAJOR_EVENTS, "In VSECSEL OpenDone");
    if (mapResolve(ps->passiveMap, 
		   &hlpType,
		   &e) == XK_FAILURE)
      {
	/* 
	 * This shouldn't happen
	 */
	xTrace0(vsecselspip, TR_ERRORS,
		"vsecselspiOpenDone: Couldn't find hlp for new session");
	return XK_FAILURE;
      }

    val = xOpenDone(e->hlpRcv, lls, self);
    xSetUp(lls, e->hlpRcv);  /* skip vsecselspi from now on */
    return	val;
}


void
vsecselspi_init( self )
    XObj self;
{
    int		i;
    int		j;
    PState	*ps;

    xTrace0(vsecselspip, TR_GROSS_EVENTS, "VSECSEL init");
    xAssert(xIsProtocol(self));
    ps = X_NEW(PState);
    self->state = (VOID *)ps;
    ps->llplist = (XObj *)xMalloc((self->numdown)*sizeof(XObj));
    ps->llptypes = (int *)xMalloc((self->numdown)*sizeof(XAlgorithm));
    ps->ip = NULL;
    ps->km = NULL;

    j = 0;

    for ( i=0; i < self->numdown; i++ ) {
	XObj	llp;

	llp = xGetDown(self, i);
	if ( ! xIsProtocol(llp) ) {
	    xTrace1(vsecselspip, TR_ERRORS, "VSECSEL can't find protocol %d", i);
	    xError("VSECSEL init errors -- not initializing");
	    return;
	} else {
	    xTrace2(vsecselspip, TR_GROSS_EVENTS, "VSECSEL llp %d is %s", i, llp->name);
	}
	if ( strcmp(llp->name,"km") == 0)
        {
		ps->km = llp;
        }
	else if ( strcmp(llp->name,"ip") == 0)
	{
		ps->ip = llp;
        }
        else 
        {
		ps->llplist[j] = llp;
		if ( llp->instName[0] == 0 )
                	ps->llptypes[j] =  relProtNumById(protTblGetId(llp->name),self);
		else
                	ps->llptypes[j] =  relProtNumById(protTblGetId(llp->instName),self);
	    	xTrace1(vsecselspip, TR_EVENTS, "VSECSEL found protocol stack %u",ps->llptypes[j]);
		j++;
	}
    }
    ps->llptypes[j] = 0;
    ps->passiveMap = mapCreate(VSECSEL_MAP_SZ, sizeof(XObj *)); 

    if (ps->km == NULL)
    { 
	    xTrace0(vsecselspip, TR_ERRORS, "VSECSEL no key manager");
	    xError("VSECSEL init errors -- no key manager");
	    return;
    }

    self->control = vsecselControlProtl;
    self->open = vsecselspiOpen;
    self->openenable = vsecselspiOpenEnable;
    self->opendisable = vsecselspiOpenDisable;
    self->opendone = vsecselspiOpenDone;
}
