/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.2
 * 
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: lance.h,v $
 *
 * HISTORY
 * $Log: lance.h,v $
 * Revision 1.3  1994/10/26 19:11:09  davidm
 * Lance descriptor is now picked up from lance_desc.h, which is shared
 * with USC.
 *
 * Revision 1.2  1994/06/11  00:02:18  menze
 * Added copyright notice
 *
 * Revision 1.1  1994/06/08  23:21:08  davidm
 * Initial revision
 *
 * History shortened.  Contributors so far: danner, af, jfriedl, rvb
 * mrt, rpd, dbg.
 */
/* 
 * Mach Operating System
 * Copyright (c) 1992,1991,1990,1989 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */
/*
 * AMD 7990 "Lance" Ethernet interface definitions.  All structures
 * are as seen from the Lance, both in the memory-alignment sense and
 * in the byte-order sense.  Mapping to host memory is model specific:
 * on pmaxen there is a 16 bit gap every other 16 bits.
 *
 * Selection of one of the four Lance CSR is done in a two-step
 * process: select which CSR first by writing into the RAP, then
 * access the register via the RDP.  Note that (a) the selection
 * remains, and (b) all but CSR0 can only be accessed when the chip is
 * stopped.  These registers are mapped in the 'registers' I/O
 * segment.
 */
#ifndef lance_h
#define lance_h

#include "xkernel.h"
#include <sys/types.h>

/*
 * A LANCE register is really 16 bits wide only.  However, the Alpha can't
 * access 16 bits in HW, so we pretend the register is 32 bits wide and
 * mask out the bottom 16 bits as necessary.
 */
typedef volatile u_int	LANCE_Register;

#define	CSR0	0x0	/* valid RAP selections */
#define	CSR1	0x1
#define	CSR2	0x2
#define	CSR3	0x3

typedef struct {
    LANCE_Register	rdp;	/* register data port */
    u_int		pad[1];
    LANCE_Register	rap;	/* register address port */
} LANCE_Registers;


/*
 * Bit definitions for the CSR0.
 * Legend:
 *	R=Readable	 W=Writeable
 *	S=Set-on-write-1 C=Clear-on-write-1
 */
#define LN_CSR0_INIT	0x0001	/* (RS) Initialize			*/
#define LN_CSR0_STRT	0x0002	/* (RS) Start				*/
#define LN_CSR0_STOP	0x0004	/* (RS) Stop				*/
#define LN_CSR0_TDMD	0x0008	/* (RS) Transmit demand			*/
#define LN_CSR0_TXON	0x0010	/* (R)  Transmitter enabled		*/
#define LN_CSR0_RXON	0x0020	/* (R)	Receiver enabled		*/
#define LN_CSR0_INEA	0x0040	/* (RW) Interrupt enable		*/
#define LN_CSR0_INTR	0x0080	/* (R)  Interrupt pending		*/
#define LN_CSR0_IDON	0x0100	/* (RC) Initialization done		*/
#define LN_CSR0_TINT	0x0200	/* (RC) Transmitter interrupt		*/
#define LN_CSR0_RINT	0x0400	/* (RC) Receiver interrupt		*/
#define LN_CSR0_MERR	0x0800	/* (RC) Memory error during DMA		*/
#define LN_CSR0_MISS	0x1000	/* (RC) No available receive buffers	*/
#define LN_CSR0_CERR	0x2000	/* (RC) Signal quality (SQE) test	*/
#define LN_CSR0_BABL	0x4000	/* (RC) Babble error: xmit too long	*/
#define LN_CSR0_ERR	0x8000	/* (R)  Error summary: any of the 4 above */

#define	LN_CSR0_WTC	0x7f00	/* Write-to-clear bits */

/*
 * Bit definitions for the CSR1.
 */
#define LN_CSR1_MBZ	0x0001	/*      Must be zero			*/
#define LN_CSR1_IADR	0xfffe	/* (RW) Initialization block address (low) */

/*
 * Bit definitions for the CSR2.
 */
#define LN_CSR2_IADR	0x00ff	/* (RW) Initialization block address (high) */
#define LN_CSR2_XXXX	0xff00	/* (RW) Reserved			*/

/*
 * Bit definitions for the CSR3.
 */
#define LN_CSR3_BCON	0x0001	/* (RW) BM/HOLD Control			*/
#define LN_CSR3_ACON	0x0002	/* (RW) ALE Control			*/
#define LN_CSR3_BSWP	0x0004	/* (RW) Byte Swap			*/
#define LN_CSR3_XXXX	0xfff8	/* (RW) Reserved			*/


/*
 * Initialization Block
 *
 * Read when the INIT command is sent to the lance.
 */
typedef struct {
    u_short	mode;			/* Mode Register, see below */
    u_short	phys_addr_high;		/* Ethernet address	*/
    u_short	phys_addr_med;		/* Ethernet address	*/
    u_short	phys_addr_low;		/* Ethernet address	*/
    u_short	logical_addr_filter0;	/* Multicast filter	*/
    u_short	logical_addr_filter1;	/* Multicast filter	*/
    u_short	logical_addr_filter2;	/* Multicast filter	*/
    u_short	logical_addr_filter3;	/* Multicast filter	*/
    u_short	recv_ring_pointer_lo;	/* Receive Ring ptr, low   */
    u_char	recv_ring_pointer_hi,	/* Receive Ring ptr, high  */
		reserved0 : 5,
		recv_ring_len : 3;	/* Length: log2(nbuffers)  */
    u_short	xmit_ring_pointer_lo;	/* Transmit Ring ptr, low  */
    u_char	xmit_ring_pointer_hi,	/* Transmit Ring ptr, high */
		reserved1 : 5,
		xmit_ring_len : 3;	/* Length: log2(nbuffers)  */
} LANCE_Init_Block;

/*
 * Bit definitions for the MODE word
 * (Normally set to 0)
 */
#define LN_MODE_DRX	0x0001	/* Disable Receiver */
#define LN_MODE_DTX	0x0002	/* Disable Transmitter */
#define LN_MODE_LOOP	0x0004	/* Loopback mode */
#define LN_MODE_DTRC	0x0008	/* Disable CRC generation */
#define LN_MODE_COLL	0x0010	/* Force collision */
#define LN_MODE_DRTY	0x0020	/* Disable retry */
#define LN_MODE_INTL	0x0040	/* Internal Loopback mode */
#define LN_MODE_XXXX	0x7f80	/* Reserved */
#define LN_MODE_PROM	0x8000	/* Promiscuous mode */

/*
 * Bit definitions for the ring pointers
 */
#define LN_RNGP_LOW	0xfffc	/* longword aligned */

/*
 * Get C version of LANCE descriptor:
 */
#include "lance_desc.h"

typedef volatile struct se_desc  *se_desc_t;

/*
 * Bit definition for STATUS byte (receive case)
 */
#define LN_RSTATE_ENP	0x01	/* (C) End of Packet			*/
#define LN_RSTATE_STP	0x02	/* (C) Start of packet			*/
#define LN_RSTATE_BUFF	0x04	/* (C) Buffer error			*/
#define LN_RSTATE_CRC	0x08	/* (C) CRC error			*/
#define LN_RSTATE_OFLO	0x10	/* (C) SILO Overflow			*/
#define LN_RSTATE_FRAM	0x20	/* (C) Framing error			*/
#define LN_RSTATE_ERR	0x40	/* (C) Error summary			*/
#define LN_RSTATE_OWN	0x80	/* (C) Owned by Lance Chip (if set)	*/

/*
 * Bit definition for STATUS byte (transmit case)
 */
#define LN_TSTATE_ENP	0x01	/* (H) End of Packet			*/
#define LN_TSTATE_STP	0x02	/* (H) Start of packet			*/
#define LN_TSTATE_DEF	0x04	/* (C) Deferred				*/
#define LN_TSTATE_ONE	0x08	/* (C) Retried exactly once		*/
#define LN_TSTATE_MORE	0x10	/* (C) Retried more than once		*/
#define LN_TSTATE_XXXX	0x20	/* Reserved				*/
#define LN_TSTATE_ERR	0x40	/* (C) Error summary (see status2)	*/
#define LN_TSTATE_OWN	0x80	/* (H) Owned by Lance Chip (if set)	*/

/*
 * Bit definitions for transmit status byte:
 */
#define LN_TXSTATE_RTRY	0x01	/* (C) Failed after 16 retransmissions	*/
#define LN_TXSTATE_LCAR	0x02	/* (C) Loss of Carrier			*/
#define LN_TXSTATE_LCOL	0x04	/* (C) Late collision			*/
#define LN_TXSTATE_XXXX	0x08	/* Reserved				*/
#define LN_TXSTATE_UFLO	0x10	/* (C) Underflow (late memory)		*/
#define LN_TXSTATE_BUFF	0x20	/* (C) Buffering error (no ENP)		*/

				/* Errors that disable the transmitter	*/
#define LN_TXSTATE_DISABLE (LN_TXSTATE_UFLO|LN_TXSTATE_BUFF|LN_TXSTATE_RTRY)

/*
 * Other chip characteristics
 */
#define LN_MINBUF_CH	100	/* Min. size of first lance buffer,
				   if chaining */

#define LN_MINBUF_NOCH	60	/* Minimum size of a lance buffer, if
				   no chaining and DTCR==1 */

#define LN_MINBUF_NOCH_RAW 64	/* Minimum size of a lance buffer, if
				   no chaining and DTCR==0 */

#endif /* lance_h */
