/*
 * $RCSfile: codegen.c,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Log: codegen.c,v $
 * Revision 1.1  1994/10/13 17:43:30  dorgival
 * Initial revision
 *
 */

/*
 * Entry points to the code generator. Makes extensive use of functions
 * in gen_utils.c
 */

#include "usc.h"
#include "symbol.h"
#include "codegen.h"
#include "y.tab.h"

static int num_statements;

typedef struct {
  int N;
  int num_icodes;
  int *array;
} RegisterArray;

static RegisterArray RegArray = {-1, 0, NULL};

void init_codegen (void)
{
  gen_opt_code = 1;
  num_statements = 0;
  set_max_key();
}

void gen_const (Buffer *buffer, Constant *theconst)
{
  switch (CONST_TYPE (theconst)) {
  case INT_CONST:
    writebuf (buffer, "(%d)", CONST_INT (theconst));
    break;

  case DOUBLE_CONST:
    writebuf (buffer, "(%lf)", CONST_DOUBLE (theconst));
    break;

  case STRING_CONST:
    writebuf (buffer, "\"%s\"", CONST_STRING (theconst));
    break;
    
  default:
    crash ("internal error in gen_const");
    break;
  }
}

int standard_type (Type *thetype)
{
  int field = TYPE_SPEC_TYPE (thetype) & ~(SIGNED_TYPE | UNSIGNED_TYPE);
  int size = TYPE_SIZE (thetype);

  if (!TYPE_IS_SPECIFIER (thetype)) {
    return 0;
  }
  if (field & ~(LONG_TYPE | INT_TYPE | SHORT_TYPE | CHAR_TYPE)) {
    return 0;
  }
  if (field == LONG_TYPE) {
    if (equal_byteorder (TYPE_BORDER (thetype), 
			 machine_info.long_info.byteorder) &&
        size == machine_info.long_info.size) {
      return 1;
    }
    else {
      return 0;
    }
  }
  else if (field == INT_TYPE) {
    if (equal_byteorder (TYPE_BORDER (thetype), 
			 machine_info.int_info.byteorder) &&
        size == machine_info.int_info.size) {
      return 1;
    }
    else {
      return 0;
    }
  }
  else if (field == SHORT_TYPE) {
    if (equal_byteorder (TYPE_BORDER (thetype), 
			 machine_info.short_info.byteorder) 
        &&
        size == machine_info.short_info.size) {
      return 1;
    }
    else {
      return 0;
    }
  }
  else if (field == CHAR_TYPE) {
    if (equal_byteorder (TYPE_BORDER (thetype), 
			 machine_info.char_info.byteorder) &&
        size == machine_info.char_info.size) {
      return 1;
    }
    else {
      return 0;
    }
  }
  else {
    crash ("internal error in standard_type");
  }
  
  return 0;
}

char *gen_offset (int x, Expr *expr)
{
  Buffer *buf = makebuf (16);
  Expr *temp;
  Expr *new;
  dprintf (500, "Entering gen_offset");
  if (x) {
    temp = int_to_expr (x);
    new = add_exprs (temp, expr);
  }
  else {
    temp = NULL;
    new = expr;
  }
  cleanbuf (buf);
  writebuf (buf, "(");
  gen_expr (buf, new, STANDARD_TYPE);
  writebuf (buf, ")");

  dprintf (700, "-----------------------------------------");
  dprintf (700, "%s", BUFFER_STRING (buf));
  dprintf (700, "%d", x);
  print_expr (700, 0, temp);
  print_expr (700, 0, expr);
  print_expr (700, 0, new);
  dprintf (700, "-----------------------------------------");
  AddToGarbage (BUFFER_STRING (buf), BUFFER_SIZE (buf));
  AddToGarbage (buf, sizeof (buf));
  dprintf (500, "Exiting gen_offset");
  return BUFFER_STRING (buf);
}

void gen_expr (Buffer *buf, Expr *expr, ExprCheckFlag expr_check_flag)
{
  int leftflag = 1, rightflag = 1;
  int priority, assoc;
  int leftpriority, leftassoc;
  int rightpriority, rightassoc;
  const char *opstring;
  Expr *leftexpr, *rightexpr;
  int leftop, rightop, op;

  if (expr) {
    switch (EXPR_TYPE (expr)) {
    case OP_EXPR:
      leftexpr = EXPR_OP_LEFT (expr);
      rightexpr = EXPR_OP_RIGHT (expr);
      op = EXPR_OP_TYPE (expr);

      opstring = parse_op (EXPR_OP_LEFT (expr),
                           op, 
                           EXPR_OP_RIGHT (expr), &priority, &assoc);

      if (!rightexpr || !EXPR_IS_OP (rightexpr) )
        rightflag = 0;
      if (!leftexpr || !EXPR_IS_OP (leftexpr) )
        leftflag = 0;

      if (leftflag) {
        leftop = EXPR_OP_TYPE (leftexpr);
        parse_op (EXPR_OP_LEFT (leftexpr), leftop,
                  EXPR_OP_RIGHT (leftexpr), &leftpriority, &leftassoc);
        if (priority > leftpriority) {
          leftflag = 1;
        }
        else if (priority == leftpriority) {
          if (assoc == RIGHT)
            leftflag = 1;
          else
            leftflag = 0;
        }
        else {
          leftflag = 0;
        }
      }

      if (rightflag) {
        rightop = EXPR_OP_TYPE (rightexpr);
        parse_op (EXPR_OP_RIGHT (rightexpr), rightop,
                  EXPR_OP_RIGHT (rightexpr), &rightpriority, &rightassoc);
        if (priority > rightpriority) {
          rightflag = 1;
        }
        else if (priority == rightpriority) {
          if (assoc == LEFT)
            rightflag = 1;
          else
            rightflag = 0;
        }
        else {
          rightflag = 0;
        }
      }

      if (leftflag)
        writebuf (buf, "(");
      gen_expr (buf, leftexpr, expr_check_flag);
      if (leftflag)
        writebuf (buf, ")");

      writebuf (buf, "%s", opstring);
      
      if (rightflag)
        writebuf (buf, "(");
      gen_expr (buf, rightexpr, expr_check_flag);
      if (rightflag)
        writebuf (buf, ")");
      break;

    case SYMINFO_EXPR:
      if (SYM_IS_SIMPLE_TYPE (EXPR_BASESYM (expr))) {
	if (  (expr_check_flag == STANDARD_TYPE )
	   && !standard_type (SYM_TYPE (EXPR_BASESYM (expr)))) {
          crash ("Can only use an unmodified type in array subscript");
	}
	else {
	  if (do_i_deref (SYM_LOCAL_NAME (EXPR_SYM (expr)))) {
	    writebuf (buf, "(*%s)", SYM_LOCAL_NAME (EXPR_SYM (expr)));
	  }
	  else {
	    writebuf (buf, "%s", SYM_LOCAL_NAME (EXPR_SYM (expr)));
	  }
	}
      }
      else {
        if ( (expr_check_flag == STANDARD_TYPE)
	   &&!standard_type (SYM_TYPE (EXPR_SYM (expr)))) {
          crash ("Can only use an unmodified type in array subscript");
        }
        else {
          if (EXPR_OFFSET (expr)) {
            writebuf (buf, "(*((%s *)(%s + %s)))", 
                      spec_type_to_string(SYM_TYPE (EXPR_SYM (expr))),
                      SYM_LOCAL_NAME (EXPR_BASESYM (expr)),
                      gen_offset (0, EXPR_OFFSET (expr)));
          }
          else {
            writebuf (buf, "(*((%s *)(%s)))", 
                      spec_type_to_string(SYM_TYPE (EXPR_SYM (expr))),
                      SYM_LOCAL_NAME (EXPR_BASESYM (expr)));
          }
        }
      }
      break;

    case CONST_EXPR:
      if (CONST_TYPE (EXPR_CONST (expr)) == INT_CONST)
	writebuf (buf, "%d", CONST_INT (EXPR_CONST (expr)));
      else
	crash ("illegal expression");
      break;

    case IDENT_EXPR:
      writebuf (buf, "%s", EXPR_IDENT (expr));
      break;

    default:
      print_expr (0, 0, expr);
      crash ("internal error in gen_expr");
      break;
    }
  }
}

int compatible_arrays (Symbol *leftsym, Symbol *rightsym, 
                       int *total_el,
                       Type **leftspec, Type **rightspec) 
{
  Type *left, *right;

  *total_el = -1;
  left = SYM_TYPE (leftsym);
  right = SYM_TYPE (rightsym);
  if (!SYM_IS_ARRAY (leftsym) || !SYM_IS_ARRAY (rightsym))
    crash ("internal error in compatible_arrays");


  if (SYM_SIZE (leftsym) != SYM_SIZE (rightsym))
    return 0;


  for ( ; left && right; 
       left = left->next, right = right->next) {
    if (TYPE_CLASS (left) == DECLARATOR &&
        TYPE_CLASS (right) == DECLARATOR) {

      if (*total_el == -1)
        *total_el = TYPE_DECL_NUMEL (left);
      else
        *total_el *= TYPE_DECL_NUMEL (left);

      if (TYPE_DECL_NUMEL (left) != TYPE_DECL_NUMEL (right))
        return 0;
     
    }
    else if (TYPE_CLASS (left) == SPECIFIER &&
             TYPE_CLASS (right) == SPECIFIER) {
      if (TYPE_IS_STRUCT_OR_UNION (left)) {
        if (!TYPE_IS_STRUCT_OR_UNION (right))
          return 0;
        if ((TYPE_IS_STRUCT (left) && !TYPE_IS_STRUCT (right)) || 
            (TYPE_IS_UNION (left) && !TYPE_IS_UNION (right)))
	  return 0;
      }
      if (TYPE_IS_STRUCT_OR_UNION (right) && !TYPE_IS_STRUCT_OR_UNION (left))
        return 0;
      *leftspec = left;
      *rightspec = right;
    }
    else if (TYPE_CLASS (left) != TYPE_CLASS (right)) {
      return 0;
    }
  }
  return 1;
}

char *add_strings (char *left, char *right)
{
  int rlen = 0, llen = 0;
  char *new;
  if (left) {
    llen = strlen (left); 
  }
  else {
    return right;
  }
  if (right) {
    rlen = strlen (right);
  }
  else {
    return left;
  }
  
  new = (char *)TempMalloc ((rlen + llen + 3) * sizeof (char));
  sprintf (new, "%s + %s", left, right);
  return new;
}

int compatible_struct_fields (Symbol *left, Symbol *right)
{
  return !strcmp (SYM_NAME (left), SYM_NAME (right));
}

void make_icode_assign (Expr *left, Expr *right)
{
  if (EXPR_IS_SYMINFO (left) && EXPR_IS_SYMINFO (right) ) {
    icode_assign (EXPR_BASESYM (left), 
		  EXPR_SYM (left), 
		  EXPR_OFFSET (left), 
		  EXPR_OFFSET (left), 
		  EXPR_BASESYM (right), 
		  EXPR_SYM (right),
		  EXPR_OFFSET (right), 
		  EXPR_OFFSET (right), 
		  NOT_IN_STRUCT);
    gen_end_icode();
    /* If lvalue is a bitfield, may need to restore other bits in that byte */

    if (SYM_IS_BITFIELD (EXPR_SYM (left))) {
      fix_bitfields(left);
    }
  }
  else if ( EXPR_IS_SYMINFO (left) && EXPR_SYM(left) ){
    icode_native_assign( SYM_LOCAL_NAME(EXPR_SYM(left)), right );
  }
  else {
    crash ("Only simple assignment expressions please");
  }
}

void make_icode_return( Expr *statement )
{
  Type* func_ret_type = SYM_FUNC_RETURN( Function );
  Type* expr_type;
  Expr* ret_expr = NULL;
  char  return_type_name[128];
  char* fmt;
  
  strcpy(return_type_name, symtype_to_string( Function ) );
  if ( ( fmt = strchr( return_type_name, '%' ) ) != NULL )
    *fmt = '\0';

  if ( !func_ret_type ) {
    if ( statement ) {
      crash("function %s should not return a value", SYM_NAME(Function));
    }
  }
  else if ( !statement ) {
    crash("function '%s' should return value of type %s\n",
	   SYM_NAME(Function), return_type_name );
  }
  else if ( TYPE_IS_POINTER( func_ret_type ) ) {
    func_ret_type = TYPE_NEXT( func_ret_type );
    if ( EXPR_IS_SYMINFO( statement ) ) {
      if ( !SYM_IS_POINTER(EXPR_SYM(statement) ) ) {
	crash("symbol is not a pointer in return statement");
      }
      else if ( func_ret_type ) {
	expr_type = TYPE_NEXT(SYM_TYPE(EXPR_SYM(statement)));
	if ( !( TYPE_IS_BASIC(func_ret_type)
	      &&( TYPE_IS_BASIC(expr_type) || standard_type(expr_type)
	      ) ) 
	   &&!( TYPE_NAME(func_ret_type) && TYPE_NAME(expr_type)
	      &&!strcmp(TYPE_NAME(func_ret_type),TYPE_NAME(expr_type))
	   )  ) {
	  crash("pointer points to something other than expect as return");
	}
      }
    }
    else if ( EXPR_IS_ADDROF( statement ) ) {
      if ( func_ret_type ) {
	if ( !EXPR_IS_SYMINFO( EXPR_OP_RIGHT( statement ) ) ) {
	  crash("internal error in make_icode_return");
	}
	expr_type = SYM_TYPE(EXPR_SYM(EXPR_OP_RIGHT( statement )));
	if ( !( TYPE_IS_BASIC(func_ret_type)
	      &&( TYPE_IS_BASIC(expr_type) || standard_type(expr_type)
	      ) ) 
	   &&!( TYPE_NAME(func_ret_type) && TYPE_NAME(expr_type)
	      &&!strcmp(TYPE_NAME(func_ret_type),TYPE_NAME(expr_type))
	   )  ) {
	  crash("pointer points to something other than expect as return");
	}
      }
    }
    else {
      crash("pointer expect as return value");
    }
  }
  else { /* returning by copy - size checked in complete_function */
    if ( !EXPR_IS_OP( statement ) ) {
      Symbol* ret_temp = ident_to_sym( SYM_NAME(Function) );
      SYM_LOCAL_NAME(ret_temp) = return_name( SYM_NAME(ret_temp) );

      ret_temp = distribute_specifier( func_ret_type, ret_temp );
      set_all_sym_type_info( ret_temp );
      ret_expr = sym_to_expr( ret_temp );
      make_icode_assign( ret_expr, statement );
    }
  }
  if ( statement )
    bit_expr_to_byte_expr( statement );

  if ( ret_expr )
    gen_temp_return_icode ( ret_expr );
  else
    gen_return_icode( statement );
}

void make_icode (Expr *statement)
{
  if (EXPR_IS_ASSIGN (statement)) {
    make_icode_assign (EXPR_OP_LEFT (statement), EXPR_OP_RIGHT (statement));
  }
  else if ( EXPR_IS_RETURN( statement ) ) {
    make_icode_return( EXPR_OP_RIGHT( statement ) );
  }
  else {
    crash ("Only simple assignment expressions please");
  }
}

char *print_offset (char *var)
{
  if (var)
    return var;
  else
    return "0";
}

void bit_to_byte (Expr *offset)
{
  if (EXPR_IS_INT (offset)) {
    EXPR_INT (offset) /= 8;
    EXPR_CONST_UNIT (offset) = BYTE_UNIT;
  }
  else if (EXPR_IS_PLUS (offset) && EXPR_IS_INT (EXPR_OP_LEFT (offset))) {
    EXPR_INT (EXPR_OP_LEFT (offset)) /= 8;
    EXPR_CONST_UNIT ((EXPR_OP_LEFT (offset))) = BYTE_UNIT;
  }
  else 
    crash ("internal error in bit_to_byte");
}

char *size_to_string (int size)
{
  if (machine_info.char_info.size == size)
    return "char";
  else if (machine_info.short_info.size == size)
    return "short";
  else if (machine_info.int_info.size == size)
    return "int";
  else if (machine_info.long_info.size == size)
    return "long";
  else
    crash ("internal error in size_to_string");
  return "";
}

Buffer *gen_left_value (ICode *icode)
{
  VarCode leftvarcode;
  char *leftstring;
  Buffer *buf;

  leftvarcode = icode->left.varcode;
  if (leftvarcode & MASK_FLAG) {
    leftstring = size_to_string (icode->left.bucket->size / 8);
  }
  else {
    leftstring = size_to_string (icode_varcode_to_size (icode->left.varcode));
  }

  if ((leftvarcode & SIGN_FLAG))
    crash ("internal error in gen_left_value");
  
  if (icode->opcode == TEMP_OP) {
    writebuf (func_vars, "const signed %s %s = ", 
              leftstring, icode->left.name);
	 write_eol (func_vars);
    buf = func_vars;
  }
  else {
    if (do_i_deref(icode->left.name)) {
      writebuf (func_body, "*((%s *)((char *)%s + %s)) =",
		leftstring, icode->left.name, 
		gen_offset (0, icode->left.bucket->start));
	   write_eol (func_body);
	   writebuf (func_body, "    ");
    }
    else {
      writebuf (func_body, "%s =", icode->left.name );
    }
    buf = func_body;
  }
  return buf;
}

RegBucket *id_to_bucket (int id)
{
  RegBucket *p;
  for (p = icode_list->right_buckets.head; p && p->bucket_id != id;
			p = p->next);
  return p;
}

RegBucket *get_right_bucket (Mask *mask)
{
  RegBucket *p;
  int id;

  if (MASK_IS_SHIFT (mask)) {
    id = MASK_BUCKET (mask);
    p = id_to_bucket (id);
  }
  else {
    p = MASK_SIGN_BUCKET (mask);
  }
  return p;
}

int not_used_again (ICode *icode, int reg)
{
  Mask *mask;
  for ( ; icode && icode->opcode == ASSIGN_OP; icode = icode->next) {
    for (mask = icode->right.mask_head; mask; mask = MASK_NEXT(mask)) {
      if (MASK_BUCKET (mask) == reg || 
	  (MASK_IS_SIGN (mask) && BUCKET_ID (MASK_SIGN_BUCKET (mask)) == reg))
	return 0;
    }
  }
  return 1;
}

int bucket_to_reg (Buffer *buf, Mask *cur_mask, ICode *cur_icode, 
						 RegBucket *bucket)
{
  int r = -1;

  for (r = 0; r < RegArray.num_icodes; r++) {
    if (RegArray.array[r] == BUCKET_ID (bucket))
      return r;
  }
  crash ("internal error in bucket_to_reg");
  return -1;
}

void gen_reg_assign (Buffer *buf, Mask *cur_mask, ICode *cur_icode,
		     RegBucket *right_bucket, int reg)
{
  int shift1;
  char *rightbucket_type = size_to_string (BUCKET_SIZE (right_bucket) / 8);
  char *reg_type = size_to_string (machine_info.reg_size / 8);

  if (BUCKET_IS_PLAIN (right_bucket)) {
    if ( do_i_deref(BUCKET_NAME (right_bucket)) ) {
    /*              ^^^^^^^^^^^^^^^^^^^^^^^^^^ was cur_icode->right.name */
      writebuf (buf, "r%d = *(unsigned %s *)((unsigned char *)%s + %s);", reg,
		rightbucket_type,
		BUCKET_NAME (right_bucket), /* was cur_icode->right.name, */
		gen_offset (0, BUCKET_START (right_bucket)));
    }
    else {
      writebuf (buf, "r%d = %s + %s;", reg,
		cur_icode->right.name,
		gen_offset (0, BUCKET_START (right_bucket)));
    }
    write_eol (buf);
  }
  else if (BUCKET_IS_SIGN (right_bucket)) {
    shift1 = machine_info.reg_size - BUCKET_SIGN_BIT (right_bucket) - 1;
    if (shift1 == 0) {
      writebuf (buf, "r%d = (signed %s)r%d >> %d;", reg, 
		reg_type,
		bucket_to_reg (buf, cur_mask, cur_icode, 
			       id_to_bucket (BUCKET_SIGN_BUCKET (right_bucket
								 ))),
		machine_info.reg_size - 1,
		BUCKET_SIGN_BUCKET (right_bucket),
		BUCKET_SIGN_BIT (right_bucket));
      write_eol (buf);
    }
    else if (shift1 > 0) {
      writebuf (buf, "r%d = (signed %s)(r%d << %d) >> %d;", reg,
		reg_type,
		bucket_to_reg (buf, cur_mask, cur_icode, 
			       id_to_bucket (BUCKET_SIGN_BUCKET (right_bucket
								 ))),
		shift1,
		machine_info.reg_size - 1,
		BUCKET_SIGN_BUCKET (right_bucket),
		BUCKET_SIGN_BIT (right_bucket));
      write_eol (buf);
    }
    else if (shift1 < 0) {
      error ("internal oddness in gen_reg_assign");
      writebuf (buf, "r%d = (signed %s)(r%d >> %d) >> %d;", reg,
		reg_type,
		bucket_to_reg (buf, cur_mask, cur_icode, 
			       id_to_bucket (BUCKET_SIGN_BUCKET (right_bucket
								 ))),
		-shift1,
		machine_info.reg_size - 1,
		BUCKET_SIGN_BUCKET (right_bucket),
		BUCKET_SIGN_BIT (right_bucket));
      write_eol (buf);
    }
  }
  else {
    crash ("internal error in gen_reg_assign");
  }
}


void allocate_reg (Mask *cur_mask, ICode *cur_icode, RegBucket *bucket)
{
  int r = -1;

  for (r = 0; r < RegArray.num_icodes; r++) {
    if (RegArray.array[r] == BUCKET_ID (bucket))
      return;
  }
  if (BUCKET_IS_SIGN (bucket)) {
    allocate_reg (cur_mask, cur_icode, 
		  id_to_bucket (BUCKET_SIGN_BUCKET (bucket)));
  }
  for (r = 0; r < RegArray.num_icodes; r++) {
    if (RegArray.array[r] < 0) {
      RegArray.array[r] = BUCKET_ID (bucket);

      writebuf (func_vars, "unsigned long r%d;", r);
      write_eol (func_vars);
      gen_reg_assign (func_body, cur_mask, cur_icode, bucket, r);
      return;
    }
    else if (not_used_again (cur_icode, RegArray.array[r])) {
      RegArray.array[r] = BUCKET_ID (bucket);
      gen_reg_assign (func_body, cur_mask, cur_icode, bucket, r);
      return;
    }
  }
  
  dprintf (0, "allocate_reg: need more registers");
  crash ("internal error in allocate_reg");

  return;
}

int gen_peek_reg (ICode *icode)
{
  Mask *mask;
  RegBucket *right_bucket;
  int any_masks = 0;

  if (icode->right.varcode & MASK_FLAG) {
    mask = icode->right.mask_head;
    while (mask) {
      if (!MASK_IS_ZERO (mask)) {
	any_masks = 1;
	right_bucket = get_right_bucket (mask);
		  
	if (right_bucket == NULL)
	  crash ("internal error in gen_peek_reg");
	allocate_reg (mask, icode, right_bucket);
      }
      mask = MASK_NEXT (mask);
    }
  }
  return any_masks;
}

int need_mask (ICode *icode, Mask *mask)
{
  if (icode->right.mask_head == NULL ||
		MASK_NEXT (icode->right.mask_head) == NULL)
	 return 0;

  if (num_set_bits (MASK_MASK (mask)) == BUCKET_SIZE (icode->left.bucket))
	 return 0;

  return 1;
}

int no_more_signs (Mask *mask)
{
  return (MASK_NEXT (mask) == NULL);
}

int highest_bit (Mask *mask)
{
  int i;
  signed long the_mask;
  
  the_mask = MASK_MASK (mask);

  for (i = sizeof (signed long) * 8 - 1; i >= 0 && the_mask >= 0; 
       i--, the_mask <<= 1);
  
  if (i < 0) {
    error ("internal warning: null mask in icode - might be a problem\n");
    return -1;
  }

  dprintf (20, "highest_bit %d\n", i - MASK_SHIFT (mask));

  return i;
}

int sign_opt_possible (ICode *icode, Mask *mask, int *lshift, int *rshift)
{
  int source_bit, lowest_sign_bit = INT_MAX;
  int highest_data_bit = -1, bucket, tmp;
  Mask *p;

  source_bit = MASK_SIGNBIT (mask);

  lowest_sign_bit = which_bit_set (MASK_MASK (mask));
  bucket = MASK_BUCKET (mask);

  for (p = icode->right.mask_head; p && p != mask; p = MASK_NEXT (p)) {
    if (MASK_BUCKET (p) == bucket) {
      tmp = highest_bit (p);
      if (tmp > highest_data_bit) {
	highest_data_bit = tmp;
      }
    }
  }

  dprintf (20, "sign_opt_possible: data %d sign %d source %d\n", 
	   highest_data_bit, lowest_sign_bit, source_bit);

  if (highest_data_bit >= 0 &&
      highest_data_bit < lowest_sign_bit) {
    *lshift = machine_info.reg_size - highest_data_bit - 1;
    *rshift = *lshift;
    return 1;
  }

  return 0;
}

void do_sign_opt (Buffer *buf, ICode *icode, Mask *mask, 
						int lshift, int rshift)
{
  dprintf (20, "sign_opt: lshift: %d; rshift: %d\n", lshift, rshift);
  writebuf (buf, " << %d >> %d", lshift, rshift);
}

void gen_right_value (Buffer *buf, ICode *icode)
{
  int reg = 0, lshift, rshift, output_in_buf = 1;
  int paren_level = 0, num_signs_seen = 0;
  VarCode rightvarcode;
  Mask *mask, *last = NULL;
  char *rightstring;
  RegBucket *right_bucket;
  static Buffer *tempbuf = NULL;

  rightvarcode = icode->right.varcode;
  rightstring = size_to_string (icode_varcode_to_size (rightvarcode));
  
  if (rightvarcode == NULL_CODE) {
    writebuf (buf, "0");
  }
  else if (rightvarcode & SIGN_FLAG & TEMP_FLAG) {
    crash ("internal error in gen_right_value");
  }
  else if (icode->right.name == NULL) {
    writebuf (buf, "%s", gen_offset (0, icode->right.offset));
  }
  else if (rightvarcode & SIGN_FLAG) {
    crash ("internal error in get_right_value");
  }
  else if (rightvarcode & TEMP_FLAG) {
    writebuf (buf, "(signed %s)%s", rightstring, icode->right.tempname);
  }
  else if (rightvarcode & MASK_FLAG) {
    output_in_buf = 0;
    if (!tempbuf)
      tempbuf = makebuf (128);
    else
      cleanbuf (tempbuf);

    writebuf (tempbuf, "(");
    paren_level++;

    mask = icode->right.mask_head;
    last = NULL;

    while (mask) {
      right_bucket = get_right_bucket (mask);
      if (!MASK_IS_ZERO (mask)) {
	if (right_bucket == NULL)
	  crash ("internal error in gen_right_value");
	reg = bucket_to_reg (tempbuf, mask, icode, right_bucket);
      }

      if (MASK_IS_SHIFT (mask)) {
	if (last && !MASK_IS_ZERO (mask)) {
	  writebuf (tempbuf, " |");
	  write_eol (tempbuf);
	  writebuf (tempbuf, "    ");
	}

	if (MASK_SHIFT(mask) == 0) {
	  if (need_mask (icode, mask))
	    writebuf (tempbuf, "(r%d & 0x%lxL)", reg, MASK_MASK(mask));
	  else
	    writebuf (tempbuf, "r%d", reg);
	}
	else if (MASK_SHIFT(mask) < 0) {
	  if (need_mask (icode, mask))
	    writebuf (tempbuf,"((r%d >> %d) & 0x%lxL)", reg, -MASK_SHIFT(mask),
		      MASK_MASK(mask));
	  else
	    writebuf (tempbuf,"(r%d >> %d)", reg, -MASK_SHIFT(mask));
	}
	else {
	  if (need_mask (icode, mask))
	    writebuf (tempbuf, "((r%d << %d) & 0x%lxL)", reg, MASK_SHIFT(mask),
		      MASK_MASK(mask));
	  else
	    writebuf (tempbuf, "(r%d << %d)", reg, MASK_SHIFT(mask));
	}
	last = mask;
      }
      else if (MASK_IS_SIGN (mask)) {
	num_signs_seen++;
	if (0 && num_signs_seen == 1 && no_more_signs (mask) && 
	    sign_opt_possible (icode, mask, &lshift, &rshift)) {
	  dprintf (50, "Doing sign opt\n");
	  writebuf (tempbuf, ")");
	  writebuf (buf, "(signed)");
	  paren_level--;
	  do_sign_opt (tempbuf, icode, mask, lshift, rshift);
	}
	else {
	  if (last) {
	    writebuf (tempbuf, " |");
	    write_eol (tempbuf);
	    writebuf (tempbuf, "    ");
	  }
			 
	  if (need_mask (icode, mask)) 
	    writebuf (tempbuf, "(r%d & 0x%lxL)", reg, MASK_MASK (mask));
	  else
	    writebuf (tempbuf, "r%d", reg);
	}
	last = mask;
      }
      else if (!MASK_IS_ZERO (mask)) {
	crash ("internal error in gen_right_value");
      }
      mask = MASK_NEXT (mask);
    }
    if (paren_level) {
      writebuf (tempbuf, ")");
      paren_level--;
    }
  }
  else {
    crash ("internal error in gen_right_value");
  }
  if (!output_in_buf)
    catbuf (buf, tempbuf); 
}

void clean_reg_array (void)
{
  int i;
  for (i = 0; i < RegArray.num_icodes; i++)
	 RegArray.array[i] = -1;
}

void new_reg_array (void)
{
  int x = 0;
  RegBucket *p;
  for (p = icode_list->right_buckets.head; p; p = p->next, x++);
  if (x > RegArray.N) {
	 if (RegArray.N >= 0)
		Free (RegArray.array, RegArray.N * sizeof (int));
	 RegArray.array = (int *)Malloc (x * sizeof (int));
	 RegArray.N = x;
  }
  RegArray.num_icodes = x;
  clean_reg_array();
}

void gen_code()
{
  ICode *icode;
  Buffer *buf;
  new_reg_array ();
  for (icode = icode_list->head; icode; icode = icode->next) {
    if (icode->opcode == PUSH_OP) {
      push_code_level();
    }
    else if (icode->opcode == FOR_OP) {
      writebuf (func_vars, "int %s;", icode->left.name);
      write_eol (func_vars);
      dont_deref (icode->left.name);
      
      writebuf (func_body, "for (%s = %s; %s < %s; %s++) ", icode->left.name, 
                gen_offset (0, icode->left.offset), icode->left.name,
                gen_offset (0, icode->right.offset), icode->left.name);
    }
    else if (icode->opcode == NAT_ASSGN_OP ) {
      writebuf(func_body,"%s = ", icode->left.name );
      gen_expr( func_body, icode->right.offset, ANY_TYPE );
      writebuf (func_body,";");
      write_eol(func_body);
    }
    else if (icode->opcode == RET_OP) {
      writebuf (func_body, "return ");
      if ( icode->right.offset ) {
	if ( icode->left.name ) {
	  writebuf(func_vars,"%s %s;",icode->left.name, icode->right.name);
	  dont_deref (icode->right.name);
	  write_eol(func_vars);
	}
	if ( EXPR_IS_ADDROF( icode->right.offset ) ) {
	  SymInfo* sym_info = EXPR_SYMINFO(EXPR_OP_RIGHT(icode->right.offset));
	  char*    ret_type_name = symtype_to_string(Function);
	  char*    tmp = strchr(ret_type_name,'%');

	  if ( tmp ) *tmp = '\0';

	  writebuf (func_body, "((%s)((char *)%s + %s))",
		    ret_type_name,
		    SYM_LOCAL_NAME(SYMINFO_BASESYM(sym_info)),
		    gen_offset (0, SYMINFO_OFFSET(sym_info)) );
	}
	else if ( EXPR_IS_SYMINFO( icode->right.offset ) )
	  gen_expr( func_body, icode->right.offset, ANY_TYPE );
	else
	  gen_expr( func_body, icode->right.offset, STANDARD_TYPE );
      }
      writebuf (func_body,";");
      write_eol(func_body);
    }
    else if (icode->opcode == POP_OP) {
      pop_code_level();
    }
    else if (icode->opcode == END_OP) {
      ;
    }
    else if (icode->opcode == ASSIGN_OP) {
      if (gen_peek_reg (icode)) {
	buf = gen_left_value (icode);
	gen_right_value (buf, icode);
	writebuf (buf, ";");
	write_eol (buf);
      }
    }
    else if (icode->opcode == TEMP_OP) {
      buf = gen_left_value (icode);
      gen_right_value (buf, icode);
      writebuf (buf, ";");
      write_eol (buf);
    }
  }
}

Expr *make_stub ( Symbol* function )
{
  char* stub_name = SYM_NAME( function );
  Expr* statement = TYPE_DECL_FBODY( SYM_TYPE( function ) );
  int   multiple_statements = ( EXPR_NEXT(statement) != NULL );

  Function = function;

  dprintf (20, "Make Stub: %s", stub_name);
  dprintf (20, "Entering make_stub: total_malloced = %d", total_malloced);
  num_statements++;
  print_expr (300, 0, statement);
  if ( CODE_IS_MACRO || ( CODE_IS_INLINE && !CODE_IS_STATIC ) )
    init_gen_utils (HeaderFile);
  else
    init_gen_utils (OutFile);
  print_source();
  temp_var_count = 0;

  icode_list = new_icodelist();

  print_expr (300, 0, statement);

  if (make_validate) {
    if (num_statements == 1) {
      init_validate();
    }
    gen_validate_code (stub_name, statement);
  }
  else {
    gen_func_header (function);
    while (statement) {

      if ( multiple_statements )
	push_code_level();

      /* loser baby. What you have to do is, after the icode_list has
	 been sorted, check if there are any bytes that have some but not
	 all bits assigned in them. If so, may need to patch things */

      init_icode();
      make_icode (statement);
      
      print_code (100, icode_list);
      optimize_icode();
      gen_code();
      statement = EXPR_NEXT(statement);
      icode_list = new_icodelist();

      if ( multiple_statements )
	pop_code_level();
    }
    gen_func_close ();
  }

  clear_deref_list();

  dprintf (100, "Exiting make_stub: total_malloced = %d", total_malloced);

  Function = NULL;
  return statement;
}
