/*
 * $RCSfile: codegen_icode.c,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Log: codegen_icode.c,v $
 * Revision 1.1  1994/10/13 17:43:30  dorgival
 * Initial revision
 *
 */

/* 
 * Intermediate code generation stuff. Traverses assignment expressions
 * checking fields and generating proper intermediate code by calling
 * functions in icode.c
 */

#include "usc.h"
#include "symbol.h"
#include "codegen.h"
#include "expr_defs.h"
#include "y.tab.h"

static int InBitHack = 0;

extern void sort_code (SortFlag which_side);


void icode_union_assign (Symbol *leftbase, 
			 Symbol *leftsym, Expr *leftoffset, 
			 Expr *real_leftoffset,
			 Symbol *rightbase, 
			 Symbol *rightsym, Expr *rightoffset,
			 Expr *real_rightoffset)
{
  if (!InBitHack)
    error ("Union assignments are ungood things");
  else
    icode_struct_assign (leftbase, leftsym, leftoffset, real_leftoffset,
			 rightbase, rightsym, rightoffset, real_rightoffset);
}



void icode_struct_assign (Symbol *leftbase, 
			  Symbol *leftsym, Expr *leftoffset, 
			  Expr *real_leftoffset,
			  Symbol *rightbase, 
			  Symbol *rightsym, Expr *rightoffset,
			  Expr *real_rightoffset)
{
  Symbol *leftfield, *rightfield;
  Expr *temp1, *temp2;
  dprintf (500, "Entering icode_struct_assign");

  leftfield = TYPE_STRUCTDEF_FIELDS (SYM_TYPE (leftsym));
  rightfield = TYPE_STRUCTDEF_FIELDS (SYM_TYPE (rightsym));

  while (leftfield && rightfield) {
    if (compatible_struct_fields (leftfield, rightfield)) {
      temp1 = bit_to_expr (SYM_OFFSET (leftfield) * 8);
      temp2 = bit_to_expr (SYM_OFFSET (rightfield) * 8);
      icode_assign (leftbase, leftfield, 
                    add_exprs (temp1, leftoffset), 
		    add_exprs (temp1, real_leftoffset),
                    rightbase, rightfield,
                    add_exprs (temp2, rightoffset),
		    add_exprs (temp2, real_rightoffset),
                    IN_STRUCT);
    }
    else {
      crash ("Incompatible types in assignment: structs differ (%s and %s)",
	      SYM_NAME(leftfield), SYM_NAME(rightfield) );
    }
    leftfield = leftfield->next;
    rightfield = rightfield->next;
  }
  if (leftfield || rightfield) {
    crash ("Incompatible types in assignment: structs differ");
  }
  dprintf (500, "Exiting icode_struct_assign");
}

void icode_basetype_assign (Symbol *leftbase, Symbol *leftsym,
			    Expr *leftoffset, Expr *real_leftoffset,
			    Symbol *rightbase, 
			    Symbol *rightsym, Expr *rightoffset,
			    Expr *real_rightoffset)
{
  int i;
  int leftsize, rightsize;
  Expr *sign_bit, *real_sign_bit, *temp1, *temp2;

  dprintf (500, "Entering icode_basetype_assign");

  leftsize = symbol_bit_size (leftsym);
  rightsize = symbol_bit_size (rightsym);

  if ((TYPE_CLASS (SYM_TYPE (leftsym)) != SPECIFIER && 
       TYPE_IS_POINTER (SYM_TYPE (leftsym))) ||
      (TYPE_CLASS (SYM_TYPE (rightsym)) != SPECIFIER &&
       TYPE_IS_POINTER (SYM_TYPE (rightsym))))
    crash ("assignment to/from pointers not allowed");

  if ((TYPE_CLASS (SYM_TYPE (leftsym)) != SPECIFIER && 
       !TYPE_IS_BITFIELD (SYM_TYPE (leftsym))) ||
      (TYPE_CLASS (SYM_TYPE (rightsym)) != SPECIFIER &&
       !TYPE_IS_BITFIELD (SYM_TYPE (rightsym))))
    crash ("internal error in icode_basetype_assign");

  if (leftsize <= rightsize) {
    for (i = 0; i < leftsize; i++) {
      temp1 = bit_to_expr (sym_bit_pos (leftsym, i, leftbase));
      temp2 = bit_to_expr (sym_bit_pos (rightsym, i, rightbase));
      gen_assign_icode (BIT_CODE, SYM_LOCAL_NAME (leftbase),
                        add_exprs (temp1, leftoffset), 
			add_exprs (temp1, real_leftoffset),
			SYM_ALIGN (leftbase), LEFT_SIDE,
                        BIT_CODE, SYM_LOCAL_NAME (rightbase),
                        add_exprs (temp2, rightoffset),
			add_exprs (temp2, real_rightoffset),
			SYM_ALIGN (rightbase), RIGHT_SIDE);
    }
  }
  else {
    for (i = 0; i < rightsize; i++) {
      temp1 = bit_to_expr (sym_bit_pos (leftsym, i, leftbase));
      temp2 = bit_to_expr (sym_bit_pos (rightsym, i, rightbase));
      gen_assign_icode (BIT_CODE, SYM_LOCAL_NAME (leftbase),
                        add_exprs (temp1, leftoffset), 
			add_exprs (temp1, real_leftoffset), 
			SYM_ALIGN (leftbase), LEFT_SIDE,
                        BIT_CODE, SYM_LOCAL_NAME (rightbase),
                        add_exprs (temp2, rightoffset), 
			add_exprs (temp2, real_rightoffset), 
			SYM_ALIGN (rightbase), RIGHT_SIDE);
    }
    if (SYM_IS_SIGNED (rightsym)) {
      sign_bit = add_exprs (bit_to_expr (sym_bit_pos (rightsym, 
						      rightsize - 1,
						      rightbase)),
			    rightoffset);
      real_sign_bit = add_exprs (bit_to_expr (sym_bit_pos (rightsym, 
							   rightsize - 1,
							   rightbase)),
				 real_rightoffset);
      for (i = 0; i < leftsize - rightsize; i++) {
	temp1 = bit_to_expr (sym_bit_pos (leftsym, i + rightsize, leftbase));
        gen_assign_icode (BIT_CODE, SYM_LOCAL_NAME (leftbase),
                          add_exprs (temp1, leftoffset), 
			  add_exprs (temp1, real_leftoffset), 
			  SYM_ALIGN (leftbase), LEFT_SIDE, 
                          BIT_CODE | SIGN_FLAG,
                          SYM_LOCAL_NAME (rightbase),
                          sign_bit, real_sign_bit, 
			  SYM_ALIGN (rightbase), RIGHT_SIDE);
      }
    }
    else {
      for (i = 0; i < leftsize - rightsize; i++) {
	temp1 = bit_to_expr (sym_bit_pos (leftsym, i + rightsize, leftbase));
	print_expr (200, 0, temp1);
        gen_assign_icode (BIT_CODE, SYM_LOCAL_NAME (leftbase),
			  add_exprs (temp1, leftoffset),
			  add_exprs (temp1, real_leftoffset),
                          SYM_ALIGN (leftbase), LEFT_SIDE, 
                          BIT_CODE | ZERO_FLAG, SYM_LOCAL_NAME (rightbase),
			  NULL, NULL,
			  SYM_ALIGN (rightbase), RIGHT_SIDE);
      }
    }
  }

  dprintf (500, "Exiting icode_basetype_assign");
}


void icode_struct_array_assign (Symbol *leftbase, Symbol *leftsym, 
				Expr *leftoffset, Expr *real_leftoffset,
				Symbol *rightbase, 
				Symbol *rightsym, Expr *rightoffset, 
				Expr *real_rightoffset,
				int total_el, Type *leftspec,
				Type *rightspec)
{
  Symbol 
    lefthack = { "", "", NULL, NULL, NULL, NULL, NULL },
    righthack = { "", "", NULL, NULL, NULL, NULL, NULL };
  Expr *temp1, *temp2;
  int i, left_steppo, right_steppo;
  Expr*   array_size_expr;
  Symbol* array_size_sym, *param;

  dprintf (500, "Entering icode_struct_array_assign: %d", total_el);

  lefthack.type = lefthack.etype = leftspec;
  righthack.type = righthack.etype = rightspec;
  lefthack.type_data = TYPE_MODIFIERS (leftspec);
  righthack.type_data = TYPE_MODIFIERS (rightspec);
  left_steppo = TYPEDATA_MEMSIZE (TYPE_MODIFIERS (leftspec)) * 8;
  right_steppo = TYPEDATA_MEMSIZE (TYPE_MODIFIERS (rightspec)) * 8;
  if (SYM_COMPLETE (leftsym)) {
    if (!SYM_COMPLETE (rightsym))
      crash ("internal error in icode_struct_array_assign");
    for (i = 0; i < total_el; i++) {
      temp1 = bit_to_expr (i * left_steppo);
      temp2 = bit_to_expr (i * right_steppo);
      icode_assign (leftbase, &lefthack, 
                    add_exprs (temp1, leftoffset),
                    add_exprs (temp1, real_leftoffset),
                    rightbase, &righthack, 
                    add_exprs (temp2, rightoffset),
                    add_exprs (temp2, real_rightoffset),
                    IN_STRUCT);
    }
  }
  else {
    if (SYM_COMPLETE (rightsym))
      crash ("internal error in icode_struct_array_assign");
    temp1 = mul_exprs (make_int_expr ("i"), 
		       bit_to_expr (left_steppo));
    temp2 = mul_exprs (make_int_expr ("i"), 
		       bit_to_expr (right_steppo));

    leftoffset = add_exprs (temp1, leftoffset);
    rightoffset = add_exprs (temp2, rightoffset);

    temp1 = leftoffset;
    temp2 = rightoffset;

    /* test to see if length was given as an int param */
    array_size_expr = (Expr*)Malloc(sizeof(Expr));
    *array_size_expr = *(TYPE_DECL_VAR_ARRAY(SYM_TYPE(rightsym)));
    if ( !EXPR_IS_IDENT( array_size_expr ) ) {
      crash("Internal error in icode_struct_array_assign: missing length");
    }

    if ( !Function ) {
      crash("Internal error in icode_struct_array_assign: missing Function");
    }

    param = TYPE_DECL_PARAMS( SYM_TYPE(Function) );

    for (array_size_sym=NULL; param; param=SYM_NEXT(param) ) {
      if ( !strcmp(SYM_NAME(param),EXPR_IDENT(array_size_expr) ) )
	break;
    }
    array_size_sym  = param;

    if ( !array_size_sym ) {
      crash ("variable length array size (%s) not declared as a parameter",
	      EXPR_IDENT (array_size_expr) );
    }

    EXPR_IDENT (array_size_expr) = local_name( EXPR_IDENT (array_size_expr) );

    gen_push_icode();
    gen_for_icode ("i", int_to_expr (0), array_size_expr );
    gen_push_icode();

    if (EXPR_IS_OP (leftoffset) || EXPR_IS_SYMINFO (leftoffset) ) {
      leftoffset = gen_temp (leftoffset);
    }
    if (EXPR_IS_OP (rightoffset) || EXPR_IS_SYMINFO (rightoffset) ) {
      rightoffset = gen_temp (rightoffset);
    }

    icode_assign (leftbase, &lefthack,
                  leftoffset, temp1,
                  rightbase, &righthack,
                  rightoffset, temp2, IN_STRUCT);

    gen_pop_icode();
    gen_pop_icode();
  }
  dprintf (500, "exiting icode_struct_array_assign");
}

void icode_assign (Symbol *leftbase, 
                   Symbol *leftsym,
		   Expr *leftoffset, Expr *real_leftoffset,
                   Symbol *rightbase, 
                   Symbol *rightsym, 
                   Expr *rightoffset, Expr *real_rightoffset,
                   AssignFlag flag)
{
  int total_el;
  Type *leftspec, *rightspec;

  dprintf (500, "Entering icode_assign");

  if (SYM_IS_STRUCT_OR_UNION (leftsym) && SYM_IS_STRUCT_OR_UNION (rightsym)) {
    if (SYM_IS_STRUCT (leftsym) && SYM_IS_STRUCT (rightsym)) {
      icode_struct_assign (leftbase, leftsym, leftoffset, real_leftoffset,
                           rightbase, rightsym, rightoffset, real_rightoffset);
    }
    else if (SYM_IS_UNION (leftsym) && SYM_IS_UNION (rightsym)) {
      icode_union_assign (leftbase, leftsym, leftoffset, real_leftoffset,
                          rightbase, rightsym, rightoffset, real_rightoffset);
    }
    else {
      crash ("Incompatible types in assignment: struct and union");
    }
  }
  else if (SYM_IS_STRUCT_OR_UNION (leftsym) || 
           SYM_IS_STRUCT_OR_UNION (rightsym)) {
    crash ("Incompatible types in assignment: struct or union "
           "and non struct or union");
  }
  else if (SYM_IS_ARRAY (leftsym) && SYM_IS_ARRAY (rightsym)) {
    if (compatible_arrays (leftsym, rightsym, &total_el, &leftspec,
			   &rightspec)) {
      icode_struct_array_assign (leftbase, leftsym, leftoffset, 
				 real_leftoffset,
				 rightbase, rightsym, rightoffset, 
				 real_rightoffset,
				 total_el, leftspec, rightspec);
    }
    else {
      crash ("Incompatible types in assignment: arrays differ");
    }
  }
  else if (SYM_IS_ARRAY (leftsym) || SYM_IS_ARRAY (rightsym)) {
    crash ("Incompatible types in assignment: array and non array");
  }
  else {
    icode_basetype_assign (leftbase, leftsym, leftoffset, real_leftoffset,
			   rightbase, rightsym, rightoffset, real_rightoffset);
  }
  dprintf (500, "Exiting icode_assign");
}

void icode_native_assign( char* var_name, Expr* expr )
{
  ICode* icode = new_icode( NAT_ASSGN_OP );
  icode->left.name = allocate_temp_string(var_name);
  icode->left.varcode = WORD_CODE;
  icode->right.offset = expr;
  add_icode (icode_list, icode);
}

void init_icode (void)
{

}

#define icode_done(a)  ((a) == NULL || (a)->opcode != ASSIGN_OP)

void icode_patch_bitfield (ICodeList *assign_icode, ICodeList *struct_icode)
{
  /* assign_icode is the real icode_list. Stick patch code into there. Use
   * struct_icode to find your way. Assuming that assign_icode has nothing
   * funny in it like loops or shit (none of that funny shit now). Just nice
   * happy little bit assignments.
   */

  ICodeList *temp_list;
  ICode *base_p, *var_p;
  ICode *new;
  int cur_byte_end;

  temp_list = new_icodelist();

  base_p = struct_icode->head;
  var_p = assign_icode->head;

  if ((base_p && base_p->opcode != ASSIGN_OP) ||
      (var_p && var_p->opcode != ASSIGN_OP)) {
    crash ("internal error in icode_patch_bitfield");
  }

  while (!icode_done (base_p) && !icode_done (var_p)) {
    cur_byte_end = (base_p->left.key & ~7) + 8;
    if (var_p->left.key >= cur_byte_end) {
      while (!icode_done (base_p) && base_p->left.key < cur_byte_end)
	base_p = base_p->next;
      continue;
    }

    while (!icode_done (base_p) && base_p->left.key < cur_byte_end) {
      if (icode_done (var_p) || base_p->left.key != var_p->left.key) {
	dprintf (20, "icode_patch_bitfield - need bit: %d", 
		 base_p->left.key);
	new = new_assign_icode (BIT_CODE, base_p->left.name, 
				base_p->left.offset, base_p->left.real_offset,
				base_p->left.alignment, LEFT_SIDE, 
				BIT_CODE, base_p->left.name,
				base_p->left.offset, base_p->left.real_offset,
				base_p->left.alignment, LEFT_SIDE);
	add_icode (temp_list, new);
	base_p = base_p->next;
      }
      else {
	var_p = var_p->next;
	base_p = base_p->next;
      }
    }
  }

  dprintf (50, "The patch code:");
  print_code (50, temp_list);

  if (temp_list->head) {
    if (assign_icode->head) {
      temp_list->tail->next = assign_icode->head;
      assign_icode->head = temp_list->head;
      assign_icode->num += temp_list->num;
    }
    else {
      assign_icode->head = temp_list->head;
      assign_icode->tail = temp_list->tail;
      assign_icode->num = temp_list->num;
    }
  }
}

void fix_bitfields (Expr *left_expr)
{
  ICodeList *temp, *struct_icode;
  Expr *offset, *struct_offset;
  Symbol *last_struct, *the_sym, *leftbasesym;

  last_struct = EXPR_LASTSTRUCT (left_expr);
  the_sym = EXPR_SYM (left_expr);
  leftbasesym = EXPR_BASESYM (left_expr);
  offset = EXPR_OFFSET (left_expr);
  struct_offset = EXPR_STRUCTOFFSET (left_expr);

  print_sym (50, 0, "sym", the_sym);
  print_expr (50, 0, offset);

  print_sym (50, 0, "struct", last_struct);
  print_expr (50, 0, struct_offset);

  sort_code (SORT_LEFT);
  temp = icode_list;   /* and just when you think it's safe to make a global
			  variable ... */

  icode_list = new_icodelist();

  /* do assign */

  offset = int_to_expr (0);
  InBitHack = 1;
  icode_assign (leftbasesym, last_struct, struct_offset, struct_offset,
		leftbasesym, last_struct, struct_offset, struct_offset, 
		NOT_IN_STRUCT);
  InBitHack = 0;

  sort_code (SORT_LEFT);

  struct_icode = icode_list;
  icode_list = temp;
  dprintf (15, "Here's the assignment:");
  print_code (15, icode_list);
  dprintf (15, "\nHere's the struct icode:");
  print_code (15, struct_icode);

  /* ok. now we've got the immediatly enclosing struct, as well as the
   * actual assignment. Now, we spit out the patchup code, using whatever
   * offset expr the struct has. Problem in that if this struct is in an
   * array, the offset we're using is the full blown expression,
   * without the temp's. This will be correct code however. 
   */

  icode_patch_bitfield (icode_list, struct_icode);

  dprintf (15, "Final assignment code:");
  print_code (15, icode_list);
}
