/*
 * $RCSfile: codegen_opt.c,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Log: codegen_opt.c,v $
 * Revision 1.1  1994/10/13 17:43:30  dorgival
 * Initial revision
 *
 */

/*
 * Intermediate code optimization
 */

#include "usc.h"
#include "symbol.h"
#include "codegen.h"
#include "bit_twiddle.h"
#include "y.tab.h"

int get_offset_key (Expr *offset)
{
  if (EXPR_IS_INT (offset))
    return EXPR_INT (offset);
  else if (EXPR_IS_PLUS (offset) && EXPR_IS_INT (EXPR_OP_LEFT (offset)))
    return EXPR_INT (EXPR_OP_LEFT (offset));
  else {
    crash ("internal error in get_offset_key");
    return -1;
  }
}

Expr *set_offset_key (Expr *offset, int key)
{
  Expr *new;
  if (EXPR_IS_CONST (offset))
    new = bit_to_expr (key);
  else
    new = add_exprs (bit_to_expr (key), EXPR_OP_RIGHT (offset)); /* was LEFT */
  return new;
}

/******************************************************************/

static ICode MAX_KEY;

void set_max_key (void)
{
  MAX_KEY.left.side = MAX_SIDE;
  MAX_KEY.right.side = MAX_SIDE;
  MAX_KEY.left.key = INT_MAX;
  MAX_KEY.right.key = INT_MAX;
}

int heapsort_compare (ICode *x, ICode *y, SortFlag which_side)
{
  IVar *x_ivar, *y_ivar;

  if (which_side == SORT_LEFT) {
    x_ivar = &(x->left);
    y_ivar = &(y->left);
  }
  else {
    x_ivar = &(x->right);
    y_ivar = &(y->right);
  }

  if (x_ivar->side < y_ivar->side) {
    return -1;
  }
  else if (x_ivar->side > y_ivar->side) {
    return 1;
  }
  else {
    if (x_ivar->key < y_ivar->key)
      return -1;
    else if (x_ivar->key > y_ivar->key)
      return 1;
    else
      return 0;
  }
}

void upheap (ICode **array, int k, SortFlag which_side)
{
  ICode *v;
  v = array[k];
  array[0] = &MAX_KEY;
  
  while (heapsort_compare (array[k / 2], v, which_side) <= 0) {
    array[k] = array[k / 2];
    k /= 2;
  }

  array[k] = v;
}

void insert (ICode **array, int *N, ICode *v, SortFlag which_side)
{
  (*N)++;
  array[*N] = v;
  upheap (array, *N, which_side);
}

void downheap (ICode **array, int N, int k, SortFlag which_side)
{
  int j;
  ICode *v;
  v = array[k];
  while (k <= N / 2) {
    j = k * 2;
    
    if (j < N && heapsort_compare (array[j], array[j + 1], which_side) < 0)
      j++;
    if (heapsort_compare (v, array[j], which_side) >= 0)
      break;

    array[k] = array[j];
    k = j;
  }
  array[k] = v;
}

ICode *remove_it (ICode **array, int *N, int k, SortFlag which_side)
{
  ICode *ans = array[1];
  array[1] = array[(*N)--];
  downheap (array, *N, 1, which_side);
  return ans;
}

ICode *sort_block (ICode **array, ICodeList *icode_list, SortFlag which_side)
{
  int k, N = 0, M;
  ICode *delim = icode_list->head;
  if (icode_list->head->opcode == ASSIGN_OP) {
    for ( ; icode_list->head && icode_list->head->opcode == ASSIGN_OP; )
      insert (array, &N, fix_icode_key (pop_icode (icode_list)), which_side);
    delim = icode_list->head;
    M = N;
    for (k = M; k >= 1; k--)
      push_icode (icode_list, remove_it (array, &N, k, which_side));
  }
  return delim;
}

/******************************************************************/

void sort_code (SortFlag which_side)
{
  ICodeList *list = (ICodeList *)Malloc (sizeof (ICodeList));
  ICode *delim = icode_list->head, **array, *temp = NULL;
  int M = icode_list->num;
  *list = *icode_list;
  
  dprintf (500, "Entering sort_code");
  print_code (500, icode_list);

  array = (ICode **)Malloc ((M + 1) * sizeof (ICode *));

  while (delim) {
    delim = sort_block (array, list, which_side);
    if (temp)
      temp->next = list->head;
    else
      icode_list->head = list->head;
    if (delim) {
      temp = delim;
      delim = delim->next;
      list->head = delim;
    }
  }
  Free ((void *)array, (M + 1) * sizeof (ICode *));
  Free ((void *)list, sizeof (ICodeList));
  dprintf (500, "Exiting sort_code");
}

void fix_code (void)
{
  ICode *p, *q = NULL;
  for (p = icode_list->head; p; p = p->next) {
    p->prev = q;
    q = p;
  }
  icode_list->tail = q;
}

void remove_mask (ICode *icode, Mask *mask)
{
  if (MASK_PREV (mask)) {
    MASK_NEXT (MASK_PREV (mask)) = MASK_NEXT (mask);
  }
  else {
    icode->right.mask_head = MASK_NEXT (mask);
  }
  if (MASK_NEXT (mask)) {
    MASK_PREV (MASK_NEXT (mask)) = MASK_PREV (mask);
  }
  else {
    icode->right.mask_tail = MASK_PREV (mask);
  }
}  

int mask_source_bit (Mask *mask)
{

  /* assuming there is one and only one source bit here */
  
  if (MASK_IS_SIGN (mask))
    return MASK_SIGNBIT (mask);
  else if (MASK_IS_SHIFT (mask))
    return which_bit_set (MASK_MASK (mask)) - MASK_SHIFT (mask);
  else
    crash ("internal error in mask_source_bit");
  return -1;
}

void append_new_bucket (int size, Expr *start, RegBucketList *b_list,
			ICodeSide bucket_side, char *name)
{
  RegBucket *end = b_list->tail;
  if (end) {
    BUCKET_NEXT (end) = (RegBucket *)TempMalloc (sizeof (RegBucket));
    b_list->tail = BUCKET_NEXT (end);
    BUCKET_ID (b_list->tail) = BUCKET_ID (end) + 1;
  }
  else {
    b_list->tail = (RegBucket *)TempMalloc (sizeof (RegBucket));
    b_list->head = b_list->tail;
    BUCKET_ID (b_list->tail) = 0;
  }
  
  BUCKET_SIDE (b_list->tail)  = bucket_side;
  BUCKET_TYPE (b_list->tail)  = PLAIN_BUCKET;
  BUCKET_SIZE (b_list->tail)  = size;
  BUCKET_START (b_list->tail) = start;
  BUCKET_NAME (b_list->tail)  = name;
  BUCKET_NEXT (b_list->tail)  = NULL;
}

RegBucket *get_sign_bucket (ICode *icode, int bucket, int bit)
{
  RegBucket *p, *orig_b = NULL;
  for (p = icode_list->right_buckets.head; p; p = BUCKET_NEXT (p)) {
    if (BUCKET_IS_SIGN (p) && BUCKET_SIGN_BUCKET (p) == bucket &&
	BUCKET_SIGN_BIT (p) == bit)
      return p;
    if (BUCKET_IS_PLAIN (p) && BUCKET_ID (p) == bucket)
      orig_b = p;
  }
  
  if (orig_b == NULL)
    crash ("internal error in get_sign_bucket");
  
  append_new_bucket (BUCKET_SIZE (orig_b), BUCKET_START (orig_b), 
		     &icode_list->right_buckets, RIGHT_SIDE, 
		     icode->right.name);
  p = icode_list->right_buckets.tail;

  BUCKET_TYPE (p)        = SIGN_BUCKET;
  BUCKET_SIGN_BUCKET (p) = bucket;
  BUCKET_SIGN_BIT (p)    = bit;
  
  return p;
}

void append_mask_to_bucket (ICode *bucket, Mask *mask)
{
  if (bucket->right.mask_head) {
    MASK_NEXT (mask) = NULL;
    MASK_PREV (mask) = bucket->right.mask_tail;
    MASK_NEXT (bucket->right.mask_tail) = mask;
    bucket->right.mask_tail = mask;
  }
  else {
    MASK_NEXT (mask) = NULL;
    MASK_PREV (mask) = NULL;
    bucket->right.mask_head = bucket->right.mask_tail = mask;
  }
}

void do_sign_optimize (ICode *icode, Mask *m1, Mask *m2)
{
  unsigned int bit = mask_source_bit (m1);
  RegBucket *sign_bucket;

  sign_bucket = get_sign_bucket (icode, MASK_BUCKET (m1), bit);

  MASK_SIGN_BUCKET (m1) = sign_bucket;
  MASK_MASK (m1) |= MASK_MASK (m2);
  SET_MASK_SIGNBIT (m1, bit);

  remove_mask (icode, m2);
}

int need_sign_optimize (Mask *m1, Mask *m2)
{
  return (((MASK_IS_SHIFT (m1) && num_set_bits (MASK_MASK (m1)) == 1) ||
	   MASK_IS_SIGN (m1)) && 
	  ((MASK_IS_SHIFT (m2) && num_set_bits (MASK_MASK (m2)) == 1) ||
	   MASK_IS_SIGN (m2)) &&
	  mask_source_bit (m1) == mask_source_bit (m2));
}

void move_signs_to_end (ICode *icode)
{
  Mask *m1, *m2, *tail;


  m1 = icode->right.mask_head;
  tail = icode->right.mask_tail;
  while (m1 != tail) {
    m2 = MASK_NEXT (m1);
    if (MASK_IS_SIGN (m1)) {
      remove_mask (icode, m1);
      append_mask_to_bucket (icode, m1);
    }
    m1 = m2;
  }
}

int sign_is_data_bit (Mask *mask)
{
  unsigned long s_bit = mask_source_bit (mask), t_bit;
  unsigned long m = MASK_MASK (mask);


  t_bit = 1L << (s_bit + MASK_SHIFT (mask));

  dprintf(15,"sign_is_data_bit: source: %d mask: 0x%x shift: %d t_bit: 0x%x\n",
	  s_bit, m, MASK_SHIFT (mask), t_bit);

  if (t_bit & m) {
    if (t_bit == m) {
      dprintf (15, "sign_is_data_bit: yes\n");
      return 1;
    }
    else {
      crash ("internal error in sign_is_data_bit");
    }
  }
  dprintf (15, "sign_is_data_bit: no\n");
  return 0;
}

void make_masks_tiny (ICode *icode)
{
  /* If N^2 is too slow you need a faster computer!! */

  Mask *m1, *m2;
  unsigned long bit;

  for (m1 = icode->right.mask_head; m1; m1 = MASK_NEXT (m1)) {
    for (m2 = MASK_NEXT (m1); m2; m2 = MASK_NEXT (m2)) {
      if (MASK_BUCKET (m1) == MASK_BUCKET (m2)) {
	if ((MASK_IS_ZERO (m1) && MASK_IS_ZERO (m2)) ||
	    (MASK_IS_SHIFT (m1) && MASK_IS_SHIFT (m2) &&
	     MASK_SHIFT (m1) == MASK_SHIFT (m2))) {
	  MASK_MASK (m1) |= MASK_MASK (m2);
	  remove_mask (icode, m2);
	}
	else if (mask_source_bit (m1) == mask_source_bit (m2) && 
		 !MASK_IS_ZERO (m1) && !MASK_IS_ZERO (m2)) {      
	  do_sign_optimize (icode, m1, m2);
	}
      }
    }
  }

  print_code (20, icode_list);

  for (m1 = icode->right.mask_head; m1; m1 = MASK_NEXT (m1)) {
    if (MASK_IS_SIGN (m1)) {
      for (m2 = icode->right.mask_head; m2; m2 = MASK_NEXT (m2)) {
	if (MASK_IS_SHIFT (m2)) {
	  if (MASK_SIGNBIT(m1) + MASK_SHIFT (m2) >= 0 && 
	      MASK_SIGNBIT (m1) + MASK_SHIFT (m2) < machine_info.reg_size)
	    bit = 1L << (MASK_SIGNBIT (m1) + MASK_SHIFT (m2));
	  else
	    bit = 0;

	  if (MASK_BUCKET (m1) == MASK_BUCKET (m2) &&
	      bit & MASK_MASK (m1)) {
	    MASK_MASK (m1) &= ~bit;
	    MASK_MASK (m2) |= bit;
	  }
	}
      }
    }
  }
  print_code (20, icode_list);
  move_signs_to_end (icode);
}

void do_optimize (void)
{
  ICode *icode;
  for (icode = icode_list->head; icode; icode = icode->next) {
    if (icode->opcode == ASSIGN_OP && (icode->left.varcode & MASK_FLAG)) {
      make_masks_tiny (icode);
    }
  }
}

int bit_in_last_bucket (ICode *icode)
{
  return (icode->prev && (icode->prev->left.varcode & MASK_FLAG) && 
	  icode->left.bucket->bucket_id == icode->prev->left.key);
}

ByteOrder bucket_order (RegBucket *bucket)
{
  ByteOrder order = NULL;
  if (bucket == NULL)
    return NULL;
  if (bucket->size / 8 == machine_info.long_info.size)
    order = machine_info.long_info.byteorder;
  else if (bucket->size / 8 == machine_info.int_info.size)
    order = machine_info.int_info.byteorder;
  else if (bucket->size / 8 == machine_info.short_info.size)
    order = machine_info.short_info.byteorder;
  else if (bucket->size / 8 == machine_info.char_info.size)
    order = machine_info.char_info.byteorder;
  else
    crash ("internal error in bucket_order");
  dprintf (100, "bucket order: %s\n", order);
  return order;
}

Mask *icode_to_mask (ICode *bit, ByteOrder left_order, ByteOrder right_order)
{
  int leftoffset, rightoffset;
  Mask *new;
  
  new = (Mask *)TempMalloc (sizeof (Mask));
  leftoffset = pos_to_bit (left_order, get_offset_key (bit->left.offset) -
			   get_offset_key (bit->left.bucket->start));

  if (ICODE_IS_ZERO (bit)) {
    SET_MASK_ZERO (new);
  }
  else {
    rightoffset = pos_to_bit (right_order, 
			      get_offset_key (bit->right.offset) - 
			      get_offset_key (bit->right.bucket->start));
   
    MASK_BUCKET (new) = bit->right.bucket->bucket_id;
    SET_MASK_SHIFT (new, leftoffset - rightoffset);
   
  }

  if (leftoffset > machine_info.reg_size)
    crash ("internal error in icode_to_mask");
  MASK_MASK (new) = 1L << leftoffset;
  return new;
}

void add_bit_to_bucket (ICode *bit, ICode *bucket)
{
  Mask *mask;

  mask = icode_to_mask (bit, bucket_order (bucket->left.bucket), 
			bucket_order (bit->right.bucket));
  append_mask_to_bucket (bucket, mask);
}

void add_new_bucket (ICode *bit)
{
  /* replace this icode with a bucket */

  Mask *mask;

  bit->left.varcode = MASK_FLAG;
  if (ICODE_IS_ZERO (bit))
    bit->right.varcode = ZERO_FLAG | MASK_FLAG;
  else
    bit->right.varcode = MASK_FLAG;
  bit->left.key = bit->left.bucket->bucket_id;

  mask = icode_to_mask (bit, bucket_order (bit->left.bucket),
			bucket_order (bit->right.bucket));
  append_mask_to_bucket (bit, mask);
}

void assign_init_masks (void)
{
  ICode *icode = icode_list->head;
  while (icode) {
    while (icode && icode->opcode != ASSIGN_OP) {
      /* just skip over this junk */
      icode = icode->next;
    }
    while (icode && icode->opcode == ASSIGN_OP) {
      if (bit_in_last_bucket (icode)) {
	dprintf (100, "bit was in last bucket\n");
	add_bit_to_bucket (icode, icode->prev);
	icode->prev->next = icode->next;
	if (icode->next)
	  icode->next->prev = icode->prev;
      }
      else {
	add_new_bucket (icode);
      }
      icode = icode->next;
    }
  }
}

Expr *zero_offset (Expr *offset)
{
  Expr *new = bit_to_expr (0);
  if (!EXPR_IS_CONST (offset)) {
    new = add_exprs (new, EXPR_OP_RIGHT (offset)); /* was RIGHT hmm */
  }
  if (EXPR_IS_CONST (new)) {
    EXPR_CONST_UNIT (new) = BIT_UNIT;
  }
  return new;
}

RegBucket *get_last_bucket (SortFlag which_side)
{
  if (which_side == SORT_LEFT) {
    return icode_list->left_buckets.tail;
  }
  else {
    return icode_list->right_buckets.tail;
  }
}

int get_max_icode (ICode *start, SortFlag which_side, ICodeSide bucket_side)
{
  ICode *p;
  IVar *the_ivar;
  int max_key = -1;

  p = start;
  while (p && p->opcode == ASSIGN_OP) {
    if (which_side == SORT_LEFT)
      the_ivar = &(p->left);
    else
      the_ivar = &(p->right);

    if (the_ivar->key > max_key && the_ivar->side == bucket_side)
      max_key = the_ivar->key;

    p = p->next;
  }
  
  dprintf (100, "get max icode: %s: %d\n", 
	   which_side == SORT_LEFT ? "left" : "right", max_key);

  return max_key;
}

int figure_size (int align, ICode *icode, 
		 int tot_size, SortFlag which_side, ICodeSide bucket_side)
{
  int size = -1;
  Expr *offset;
  int delta;

  tot_size = round_up_byte (tot_size) / 8;

  if (which_side == SORT_LEFT) 
    offset = icode->left.real_offset;
  else
    offset = icode->right.real_offset;

  delta = get_offset_key (offset) / 8;

  if (tot_size <= -8)
    crash ("internal error in figure_size");

  if (tot_size >= machine_info.long_info.size &&
      aligns_match (delta, machine_info.long_info.alignment) &&
      aligns_match (align, machine_info.long_info.alignment))
    size = machine_info.long_info.size * 8;
  else if (tot_size >= machine_info.int_info.size &&
	   aligns_match (delta, machine_info.int_info.alignment) &&
	   aligns_match (align, machine_info.int_info.alignment))
    size = machine_info.int_info.size * 8;
  else if (tot_size >= machine_info.short_info.size &&
	   aligns_match (delta, machine_info.short_info.alignment) &&
	   aligns_match (align, machine_info.short_info.alignment))
    size = machine_info.short_info.size * 8;
  else if (tot_size >= machine_info.char_info.size &&
	   aligns_match (delta, machine_info.char_info.alignment) &&
	   aligns_match (align, machine_info.char_info.alignment))
    size = machine_info.char_info.size * 8;
  else 
    crash ("internal error in figure_size");
  
  return size;
}

int icode_is_not_bckt_delim (ICode *p, SortFlag which_side, 
			     ICodeSide bucket_side)
{
  IVar *s;

  if (which_side == SORT_LEFT)
    s = &(p->left);
  else
    s = &(p->right);

  return (p && p->opcode == ASSIGN_OP && s->side == bucket_side);
}

char *get_icode_name (ICode *p, SortFlag which_side)
{
  if (which_side == SORT_LEFT)
    return p->left.name;
  else
    return p->right.name;
}

ICode *bucket_block (int alignment, ICode *start,
		     RegBucketList *b_list, 
		     SortFlag which_side, ICodeSide bucket_side)
{
  ICode *delim;
  int cur_end, cur_size;
  Expr *cur_start;
  int done;
  int tot_size;

  if (which_side == SORT_RIGHT) {
    while (start && ICODE_IS_ZERO (start))
      start = start->next;
  }

  delim = start;

  if (icode_is_not_bckt_delim (delim, which_side, bucket_side)) {
    done = 0;
   
    tot_size = get_max_icode (start, which_side, bucket_side) + 1;;

    cur_size = figure_size (alignment, start, tot_size, 
			    which_side, bucket_side);
    tot_size -= cur_size;

    if (which_side == SORT_LEFT)
      cur_start = zero_offset (delim->left.offset);
    else
      cur_start = zero_offset (delim->right.offset);
    cur_end = cur_size;
    append_new_bucket (cur_size, cur_start, b_list, bucket_side,
		       get_icode_name (delim, which_side));

    while (!done) {
      for ( ; icode_is_not_bckt_delim (delim, which_side, bucket_side) && 
	   icode_key (delim, which_side) < cur_end ; delim = delim->next) {
	if (which_side == SORT_LEFT || !ICODE_IS_ZERO (delim)) {
	  if (which_side == SORT_LEFT) {
	    delim->left.bucket = b_list->tail;
	  }
	  else {
	    delim->right.bucket = b_list->tail;
	  }
	}
      }

      /* if there is no more code left in this assignment, then use a smaller
       * bucket size. else, just read/write over it. if something else
       * is supposed to be there, it will get written later.
       */

      if (which_side == SORT_RIGHT) {
	while (delim && ICODE_IS_ZERO (delim))
	  delim = delim->next;
      }
      if (icode_is_not_bckt_delim (delim, which_side, bucket_side)) {
	cur_size = figure_size (alignment, start, tot_size, 
				which_side, bucket_side);
	tot_size -= cur_size;
	cur_start = set_offset_key (cur_start, cur_end);
	cur_end += cur_size;
	append_new_bucket (cur_size, cur_start, b_list, bucket_side,
			   get_icode_name (delim, which_side));
      }
      else {
	done = 1;
      }
    }
   
  }
  return delim;
}

void make_buckets (void)
{
  int alignment;
  ICode *delim = icode_list->head;

  print_code (30, icode_list);
  sort_code(SORT_RIGHT);
  fix_code();
  dprintf (30, "Before bucket assign\n");
  print_code (30, icode_list);
  

  alignment = -1;
  delim = icode_list->head;

  while (delim) {
    while (delim && delim->opcode != ASSIGN_OP) {
      delim = delim->next;
    }
    while (delim && delim->opcode == ASSIGN_OP) {
      if (alignment < 0)
	alignment = delim->right.alignment;
      delim = bucket_block (alignment, delim, &icode_list->right_buckets, 
			    SORT_RIGHT, LEFT_SIDE);
      delim = bucket_block (alignment, delim, &icode_list->right_buckets, 
			    SORT_RIGHT, RIGHT_SIDE);
      delim = delim->next; 
    }
  }
  dprintf (30, "After bucket assign\n");
  print_code (30, icode_list);
  sort_code(SORT_LEFT);
  fix_code();
  print_code (30, icode_list);

  alignment = -1;
  delim = icode_list->head;

  while (delim) {
    while (delim && delim->opcode != ASSIGN_OP) {
      delim = delim->next;
    }
    while (delim && delim->opcode == ASSIGN_OP) {
      if (alignment < 0)
	alignment = delim->left.alignment;
      delim = bucket_block (alignment, delim, &icode_list->left_buckets,
			    SORT_LEFT, LEFT_SIDE);
      delim = delim->next; 
    }
  }
  print_buckets(30);
  print_code (30, icode_list);
}

void byte_buckets (void)
{
  RegBucket *p;
  for (p = icode_list->left_buckets.head; p; p = p->next) {
    if (BUCKET_IS_PLAIN (p))
      bit_expr_to_byte_expr (p->start);
  }
  for (p = icode_list->right_buckets.head; p; p = p->next) {
    if (BUCKET_IS_PLAIN (p))
      bit_expr_to_byte_expr (p->start);
  }
}

void convert_bits_to_bytes (void)
{
  /* need to convert over all buckets, and temps */

  ICode *p;

  byte_buckets();

  for (p = icode_list->head; p; p = p->next) {
    if (p->opcode == TEMP_OP) {
      bit_expr_to_byte_expr (p->left.offset);
      bit_expr_to_byte_expr (p->right.offset);
    }
  }
}

void optimize_icode (void)
{    
  dprintf (50, "Entering optimize_code");

  make_buckets();       
  assign_init_masks();
  do_optimize();

  print_buckets (50);
  print_code (50, icode_list);

  convert_bits_to_bytes();
  print_buckets (20);
  print_code (20, icode_list);
  dprintf (50, "Exiting optimize_code");
}
