/*
 * $RCSfile: gen_utils.c,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Log: gen_utils.c,v $
 * Revision 1.3  1994/10/30 20:32:08  dorgival
 * Fixed return type for time() to work in the Alpha (in init_outfile)
 *
 * Revision 1.2  1994/10/13  23:48:42  dorgival
 * Added "#include <ctype.h>" - toupper may be a macro there...
 *
 * Revision 1.1  1994/10/13  17:43:30  dorgival
 * Initial revision
 *
 */

/*
 * Miscelanous functions, most of them related to output generation
 */

#include "usc.h"
#include "symbol.h"
#include "gen_utils.h"
#include <ctype.h>

#ifdef NO_STRFTIME
#define strftime(string,maxsize,format,timeptr) \
        strlen(strcpy(string,asctime(timeptr)))
#endif

extern char *spec_type_to_string (Type* type);

CodeLevel *code_stack;
NameList *ParamList, *DontDerefList;
FILE *CodeOutFile;

Buffer *func_header = NULL, *func_body, *func_vars;
int code_stack_depth;

static int param_flag;

int pos_to_bit (ByteOrder order, int pos)
{
  int bit;
  bit = pos / 8;
  bit = order[bit] - 1;
  bit = bit * 8 + (7 - pos % 8);
  dprintf (20, "pos_to_bit: order: %s pos: %d bit: %d",
		byteorder_to_string(order), pos, bit);
  return bit;
}

int byte_pos (int byte, ByteOrder order, int memsize )
{
  int temp = ((byte - 1) % memsize ) + 1;
  int len  = byteorder_len (order);
  int i, b = 0, used_bytes[MAX_BORDER_SIZE];

  dprintf (500, "Entering byte_pos: %d %s %d",
		 byte, byteorder_to_string(order), memsize);

  if ( (temp < 1)||(temp>len)||(memsize>MAX_BORDER_SIZE) )
    crash("Internal error in byte_pos: requested byte %d, memsize %d",
	   temp, memsize );

  memset( (char *)used_bytes, 0, sizeof(used_bytes) );
  for (i=0; i<MAX_BORDER_SIZE && order[i]; i++) {
    b = order[i]-1;
    used_bytes[ b ] = i+1;
  }

  for (i=0;i<MAX_BORDER_SIZE;i++)
    if ( used_bytes[i] == temp )
      b = i;
  if ( b==MAX_BORDER_SIZE )
    crash("Internal error in byte_pos");

  dprintf (500, "Exiting byte_pos: %d\n", b);
  return byte - temp + b;
}

int sym_bit_pos (Symbol *sym, int bit, Symbol *base_sym)
{
  int pos, memsize;
  BitInfo *bits;

  dprintf(500,"Entering sym_bit_pos");
  if (SYM_IS_BITFIELD (sym)) {
    bits = TYPE_DECL_BITINFO(SYM_TYPE (sym));
    bit = bits[bit];
  }

  if ( SYM_IS_ARRAY(base_sym) ) {
    memsize = TYPE_MEMSIZE( SYM_TYPE(sym) );
  } 
  else {
    memsize = SYM_MEMSIZE(sym);
  }

  pos = byte_pos ((bit / 8) + 1, SYM_BORDER (sym), memsize );
  pos = pos * 8 + (7 - bit % 8);

  dprintf (500, "bit_pos: %d -> %d\n", bit, pos);
  dprintf(500,"Exiting sym_bit_pos");
  return pos;
}

char *symtype_to_string (Symbol *sym)
{
  Type* type;
  static char s[128];
  char tmp[128];
  s[0] = '\0';
  tmp[0] = '\0';
#define insert_name( x )  strcat( (x)," %s")

  if ( SYM_IS_FUNCTION(sym) )
    type = SYM_FUNC_RETURN(sym);
  else
    type = SYM_TYPE(sym);

  if ( type == NULL )
    return "void %s";

  if ( TYPE_NAME(type) ) {
    strcpy(s,TYPE_NAME(type));
    insert_name(s);
    return s;
  }

  if (!TYPE_IS_SPECIFIER (type))
    if ( TYPE_IS_POINTER(type) ) {
      strcpy(s, spec_type_to_string( TYPE_NEXT(type) ) );
      strcat(s,"*");
      insert_name(s);
      return s;
    }
    else if ( TYPE_IS_FUNCTION(type) ) {
      strcpy(s, spec_type_to_string( TYPE_NEXT(type) ) );
      insert_name(s);
      return s;
    }
    else if ( TYPE_IS_ARRAY(type) ) {
      while ( type && TYPE_IS_ARRAY(type) ) {
	if ( TYPE_DECL_NUMEL(type) > 0 ) {
	  sprintf( tmp, "[%d]", TYPE_DECL_NUMEL(type) );
	  strcat( s, tmp );
	}
	else
	  strcat( s, "[]" );
	type = TYPE_NEXT( type );
      }
      strcpy(tmp,s);
      strcpy(s,spec_type_to_string( type ) );
      insert_name(s);
      strcat(s,tmp);
      return s;
    }
    else
      crash ("internal error in symtype_to_string: not a specifier. bad bad.");
  return spec_type_to_string(type);
}

int compatible_return_type( Symbol* function, Symbol* sym )
{
  char function_type_name[128];
  char sym_type_name[128];
  char* fmt;

  strcpy(function_type_name, symtype_to_string(function) );
  if ( ( fmt = strchr( function_type_name, '%' ) ) != NULL )
    *fmt = '\0';
  strcpy(sym_type_name, symtype_to_string(sym) );
  if ( ( fmt = strchr( sym_type_name, '%' ) ) != NULL )
    *fmt = '\0';

  if ( strncmp( function_type_name,"void *",sizeof("void *") ) )
    return (strcmp(function_type_name,sym_type_name) == 0);
  else
    return (strchr(sym_type_name,'*') != NULL );
}

void init_gen_utils (FILE *outfile)
{
  code_stack = NULL;
  code_stack_depth = 0;
  ParamList = NULL;
  CodeOutFile = outfile;
  if (func_header == NULL)
    func_header = makebuf (FUNCHEADERBUFFER);
  else
    cleanbuf (func_header);
}

void init_outfile (OutputType output_type)
{  
  char string[64];
# ifdef NO_TIME_T
    long time1 = time(NULL);
# else
    time_t time1 = time(NULL);
# endif
  struct tm *time2 = localtime (&time1);
  int count, n, i;
  char* lfptr;
  FILE* prev_code_out_file = NULL;

  count = strftime (string, 64, "%I:%M:%S %p %b %d, %Y", time2);
  lfptr = strchr(string,'\n');
  if ( lfptr ) {
    *lfptr = '\0';
    count = strlen(string);
  }

  switch( output_type ) {
  case USC_OUTPUT:
  case USIT_OUTPUT:
    init_gen_utils (OutFile);
    break;
  case USC_TYPES:
    prev_code_out_file = CodeOutFile;
    CodeOutFile = HeaderFile;
    break;
  }

  genout ("/**********************************************************/\n");
  genout ("/*                                                        */\n");
  switch( output_type ) {
    case USC_OUTPUT:
      genout ("/*                  USC Output                            */\n");
      break;
    case USIT_OUTPUT:
      genout ("/*                  USIT Output                           */\n");
      break;
    case USC_TYPES:
      genout ("/*                  USC Types                             */\n");
      break;
  }

  genout ("/*");
  n = 56 - count;
  for (i = 0; i < 18; i++)
    genout (" ");
  genout ("%s", string);
  for (; i < n; i++)
    genout (" ");
  genout ("*/\n");
  genout ("/*                  File: ");
  
  if (strcmp (InFileName, "-")) {
    count = strlen (InFileName);
    genout (InFileName);
  }
  else {
    count = strlen ("Standard Input");
    genout ("Standard Input");
  }
  for (i = 0; i < 32 - count; i++)
    genout (" ");
  genout("*/\n");
  genout ("/*                                                        */\n");
  genout ("/**********************************************************/\n\n");
  if ( ( output_type == USC_OUTPUT ) && preserve_type_names) {
    genout ("#define __USC_STUBS__\n");
  }
  if ( output_type == USC_TYPES ) {
    CodeOutFile = prev_code_out_file;
  }
}

void clear_deref_list(void)
{
  DontDerefList = NULL;
}

void write_eol (Buffer *buf)
{
  if (CODE_IS_MACRO)
    writebuf (buf, "\\\n");
  else
    writebuf (buf, "\n");
}

void gen_eol (void)
{
  if (CODE_IS_MACRO)
    fprintf (CodeOutFile, "\\\n");
  else
    fprintf (CodeOutFile, "\n");
}

void genout (char *fmt, ...)
{
  va_list args;

  va_start(args, fmt);
  vfprintf (CodeOutFile, fmt, args);
  va_end (args);
}

void push_code_level (void)
{
  CodeLevel *new = (CodeLevel *)TempMalloc (sizeof (CodeLevel));
  code_stack_depth++;
  func_body = new->func_body = makebuf (FUNCBODYBUFFER);
  func_vars = new->func_vars = makebuf (FUNCVARBUFFER);
  new->next = code_stack;
  code_stack = new;
  write_eol (func_vars);

  if (code_stack_depth == 1 && CODE_IS_MACRO)
    writebuf (func_vars, "do {");
  else
    writebuf (func_vars, "{");
  write_eol (func_vars);
}

void genbuf (Buffer *buffer)
{
  fprintf (CodeOutFile, "%s", BUFFER_STRING (buffer));
}

void pop_code_level (void)
{
  code_stack_depth--;
  if (!code_stack)
    crash ("internal error in pop_code_level");
  if (code_stack->next) {
    writebuf (func_body, "}");
	 write_eol (func_body);
    catbuf (code_stack->next->func_body, code_stack->func_vars);
    catbuf (code_stack->next->func_body, code_stack->func_body);
    freebuf (code_stack->func_body);
    freebuf (code_stack->func_vars);
    code_stack = code_stack->next;
    func_vars = code_stack->func_vars;
    func_body = code_stack->func_body;
  }
  else {
    genbuf (code_stack->func_vars);
    genbuf (code_stack->func_body);
    freebuf (code_stack->func_body);
    freebuf (code_stack->func_vars);
    genout ("}");
    gen_eol ();
    code_stack = code_stack->next;
  }
}

void gen_func_name_header (char *type, char *ident)
{
  char* static_decl;
  cleanbuf (func_header);
  param_flag = 0;

  push_code_level();
  static_decl = (CODE_IS_STATIC) ? "static " : "";
  if (CODE_IS_INLINE)
    writebuf (func_header, "%sinline %s %s ()", static_decl, type, ident);
  else if (CODE_IS_PLAIN)
    writebuf (func_header, "%s%s %s ()", static_decl, type, ident);
  else if (CODE_IS_MACRO)
    writebuf (func_header, "#define %s()", ident);
  else
    crash ("internal error in gen_func_name_header");
}

void gen_func_header (Symbol* function)
{
  char* ident = SYM_NAME(function);
  char  return_type[128];
  Symbol* param = TYPE_DECL_PARAMS( SYM_TYPE(function) );
  char* fmt;

  strcpy(return_type,symtype_to_string( function ));
  if ( ( fmt = strchr( return_type, '%' ) ) != NULL )
    *fmt = '\0';

  strcat(return_type," %s(");
  cleanbuf (func_header);
  param_flag = 0;
  push_code_level();

  if ( CODE_IS_MACRO ) {
    if ( SYM_FUNC_RETURN(function) )
      error("%s must return void to be defined as macro", ident);
    writebuf (func_header, "#define %s(", ident);
  }
  else {
    if ( CODE_IS_STATIC )
      writebuf (func_header,"static ");
    if ( CODE_IS_INLINE )
      writebuf (func_header, "inline ");
    writebuf (func_header, return_type, ident);
  }

  while (param) {
    add_param( 1, param, NORMAL_PARAM );
    param = SYM_NEXT(param);
  }

  writebuf (func_header, ")");
  if ( HeaderFile != OutFile && CODE_IS_PLAIN && !CODE_IS_STATIC )
    fprintf (HeaderFile, "%s;\n\n", BUFFER_STRING (func_header));

}

void gen_func_close (void)
{
  genbuf (func_header);
  pop_code_level();

  if (CODE_IS_MACRO)
    genout (" while (0)\n\n");
  else
    genout ("\n\n");
  CleanGarbage();
}

int added_param (char *name)
{
  NameList *p = ParamList;
  int found = 0;
  while (p && !found) {
    if (!strcmp (name, p->name))
      found = 1;
    p = p->next;
  }
  return found;
}

void add_param_name (char *name)
{
  NameList *new = (NameList *)TempMalloc (sizeof (NameList));
  new->name = name;
  new->next = ParamList;
  ParamList = new;
}


void add_param (int n, Symbol *sym, ParamFlag flag, ...)
{
  va_list ap;
  va_start (ap, flag);


  while (n > 0) {
    if (!added_param (SYM_LOCAL_NAME (sym))) {
      add_param_name (SYM_LOCAL_NAME (sym));
      if (param_flag) {
	if (!CODE_IS_MACRO)
	  writebuf (func_header, ", ");
	else
	  writebuf (func_header, ",");
      }
      param_flag = 1;
      
      if (flag == NORMAL_PARAM) {
	if (!CODE_IS_MACRO){
	    char* param = symtype_to_string (sym);
	    if ( ! preserve_type_names && (strpbrk(param,"*[")!=NULL) ) {
	      param = "void* %s";
	    }
	    if ( strchr(param,'%') == NULL )
	      writebuf (func_header, "%s %s", param, SYM_LOCAL_NAME (sym));
	    else
	      writebuf (func_header, param, SYM_LOCAL_NAME (sym) );
	}
	else
	  writebuf (func_header, "%s", SYM_LOCAL_NAME (sym));
      }
      else {
	if (!CODE_IS_MACRO)
	  writebuf (func_header, "void *%s", SYM_LOCAL_NAME (sym));
	else
	  writebuf (func_header, "%s", SYM_LOCAL_NAME (sym));
      }
    }
    n--;
    if (n > 0) {
      sym = va_arg (ap, Symbol *);
      flag = va_arg (ap, ParamFlag);
    }
  }
}   

void dont_deref (char *name)
{
  NameList *new = (NameList *)TempMalloc (sizeof (NameList));

  new->name = name;
  new->next = DontDerefList;
  DontDerefList = new;
}

int do_i_deref (char *name)
{
  NameList *p = DontDerefList;

  int found = 0;
  while (p && !found) {
    if (!strcmp (name, p->name))
      found = 1;
    p = p->next;
  }
  return !found;
}

void print_place_holder( int num_tabs, Symbol* sym, FILE* out )
{
  char* indent = make_indent(num_tabs);
  SpecifierTypeField specfield;
  Type etype = *(SYM_ETYPE (sym));

  TYPE_NAME(&etype) = NULL;

  if (!TYPE_IS_SPECIFIER (SYM_ETYPE(sym)))
    crash ("internal error in print_place_holder");
  specfield = TYPE_SPEC_TYPE (SYM_ETYPE (sym));
  fprintf (out, "%s%s", indent, spec_type_to_string (&etype));
  if (specfield & (UNION_TYPE | STRUCT_TYPE)) {
    fprintf (out, "{\n");
    if ( !SYM_TYPEDATA(sym) )
      crash( "symbol has no typedata in print_place_holder");
    fprintf (out, "  %schar filling[%d];\n", indent, SYM_MEMSIZE(sym) );
    fprintf (out, "%s} ", indent);
  }
}

void write_fake_type( FILE* out, Symbol* sym )
{
  char*   sym_name;
  Type*   sym_type;
  char    macro_name[128];
  int     i;

  while (sym) {
    sym_name = SYM_NAME(sym);
    sym_type = SYM_TYPE(sym);

    for (i=0;sym_name[i];++i)
      macro_name[i] = toupper(sym_name[i]);
    macro_name[i] = '\0';

    fprintf( out, "#if ( defined(__USC_STUBS__) || defined(USE_USC_%s) )\n",
		  macro_name );

    fprintf( out, "typedef ");

    print_place_holder( 0, sym, out );

    fprintf( out, "%s", sym_name );

    while (sym_type && !TYPE_IS_SPECIFIER(sym_type) ) {
      if ( TYPE_IS_ARRAY( sym_type ) ) {
	if ( TYPE_DECL_NUMEL( sym_type ) >=0 ) 
	  fprintf( out, "[%d]", TYPE_DECL_NUMEL( sym_type ));
	else
	  fprintf( out, "[0]");
      }
      sym_type = TYPE_NEXT(sym_type);
    }

    fprintf( out, ";\n" );

    fprintf( out, "#endif /* %s */\n\n", macro_name );

    sym = SYM_NEXT(sym);
  }
}
